import os
from collections import OrderedDict
from typing import Union

import astropy.units as u
import numpy as np
from astropy.table import QTable, hstack, vstack

import exosim.log as log
import exosim.recipes as recipes
import exosim.tasks.radiometric as radiometric
from exosim.output import SetOutput
from exosim.utils.ascii_arts import astronomer4
from exosim.utils.klass_factory import find_task
from exosim.utils.prepare_recipes import clean_config_files, load_options
from exosim.utils.runConfig import RunConfig
from exosim.utils.timed_class import TimedClass


class RadiometricModel(TimedClass, log.Logger):
    """

    Pipeline to create the radiometric model.
    This pipeline has three working modes:

    - it can load an already produced focal plane and use it to estimate a radiometric model;
    - it can produce a single sourec focal planet and estimate the radiometric model;
    - it can load a target list and produce the radiometric model for each target of the target list.

    Attributes
    ------------
    mainConfig: dict
        This is parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
    input: :class:`~exosim.output.output.Output`
        input/output file
    payloadConfig: dict
        payload configuration dictionary extracted from mainConfig`
    table: :class:`~astropy.table.QTable`
        table for the radiometric estimations

    Examples
    --------

    If the user wants to estimate the radiometric model af an existing focal plane

    >>> import exosim.recipes as recipes
    >>> rm = recipes.RadiometricModel(options_file= 'main _configuration.xml',
    >>>                                input_file = 'focal_plane.h5')

    Otherwhise, if a focal planet has not been produced yet, this recipy can produce it,
    if a destination not existing file is provided:

    >>> import exosim.recipes as recipes
    >>> rm = recipes.RadiometricModel(options_file= 'main _configuration.xml',
    >>>                                input_file = 'desired_output.h5')

    In both cases, to store the produced table into the output file,
    the :func:`~exosim.recipes.radiometricModel.RadiometricModel.write`
    is to be used:

    >>> rm.write()

    """

    def __init__(
        self, options_file: Union[str, dict], input_file: str
    ) -> None:
        """

        Parameters
        ----------
        options_file: str or dict
            input configuration file
        input_file: str
            input file
        """
        super().__init__()

        self.graphics(astronomer4)
        RunConfig.stats()
        self.announce("started")
        clean_config_files()
        self.mainConfig, self.payloadConfig = load_options(options_file)

        self.table = None

        # if focal plane already exists it loads it
        if os.path.isfile(input_file):
            self.input = SetOutput(input_file, replace=False)
            self.info("Running radiometric model on existing focal plane")
            self.single_file_pipeline()
        # if focal plane doesn't exist it create it for a single step
        elif input_file.endswith("h5"):
            self.info(
                "Focal plane not found: running CreateFocalPlane pipeline"
            )
            # single time step
            self.mainConfig["time_grid"] = {
                "start_time": 0 * u.hr,
                "end_time": 1 * u.hr,
                "low_frequencies_resolution": 1 * u.hr,
            }
            self._isolate_every_opt()
            recipes.CreateFocalPlane(
                options_file=self.mainConfig, output_file=input_file
            )
            self.input = SetOutput(input_file, replace=False)
            self.single_file_pipeline()

        # if input is a target list, it creates the focal plane only for
        # instruments and foregrounds
        elif input_file.endswith(".csv"):
            raise NotImplementedError
            # # single time step
            # self.mainConfig['time_grid'] = {'start_time': 0 * u.hr,
            #                                 'end_time': 1 * u.hr,
            #                                 'low_frequencies_resolution': 1 * u.hr}
            # # remove source
            # if 'source' in self.mainConfig['sky'].keys():
            #     self.mainConfig['sky'].pop('source')
            # self._isolate_every_opt()
            # recipes.CreateFocalPlane(options_file=self.mainConfig,
            #                          output_file=input_file)

        self.common_pipeline()
        self.write()
        self.log_runtime_complete("recipe ended", "info")
        self.announce("ended")

    def _isolate_every_opt(self) -> None:
        """
        it iterates over the optical elements to isolate them and craete sub
        focal planes
        """
        from exosim.utils.iterators import iterate_over_opticalElements

        self.mainConfig["sky"] = iterate_over_opticalElements(
            self.mainConfig["sky"], "foregrounds", "isolate", True
        )
        self.payloadConfig = iterate_over_opticalElements(
            self.payloadConfig, "Telescope", "isolate", True
        )
        self.payloadConfig = iterate_over_opticalElements(
            self.payloadConfig, "channel", "isolate", True
        )

    def single_file_pipeline(self) -> None:
        """
        Radiometric pipeline to run for a single target with an already
        produced focal plane. The involved steps are:

        1. creation of the wavelength table with :func:`~exosim.recipes.radiometricModel.RadiometricModel.create_table`;
        2. estimation of the apertures sizes and number of pixels involved with :func:`~exosim.recipes.radiometricModel.RadiometricModel.compute_apertures`;
        3. estimation of the signals in the apertures for the sub foregrounds, if any: :func:`~exosim.recipes.radiometricModel.RadiometricModel.compute_sub_foregrounds_signals`;
        4. estimation of the total foreground signal in the apertures:  :func:`~exosim.recipes.radiometricModel.RadiometricModel.compute_foreground_signals`;
        5. estimation of the source focal plane signal in the aperture: :func:`~exosim.recipes.radiometricModel.RadiometricModel.compute_source_signals`;
        6. estimation of the saturation time in the channel: :func:`~exosim.recipes.radiometricModel.RadiometricModel.compute_saturation`;

        The pipeline will update the `table` attribute.
        """
        self.table = self.create_table()
        self.compute_apertures()
        self.compute_sub_foregrounds_signals()
        self.compute_foreground_signals()
        self.compute_source_signals()
        self.compute_saturation()

    def common_pipeline(self) -> None:
        """
        Radiometric pipeline to run starting from a radiometric table with already estimated signals.
        It computes the noise.

        1. estimation of the multiaccum factors :func:`~exosim.recipes.radiometricModel.RadiometricModel.compute_multiaccum`;
        2. estimation shot noise :func:`~exosim.recipes.radiometricModel.RadiometricModel.compute_photon_noise`;
        3. update total noise :func:`~exosim.recipes.radiometricModel.RadiometricModel.update_total_noise`

        The pipeline will update the `table` attribute.
        """
        self.compute_multiaccum()
        self.compute_photon_noise()
        self.update_total_noise()

    def create_table(self) -> QTable:
        """
        Produces the starting radiometric table with the spectral bins and their edges.
        It is based on :class:`~exosim.tasks.radiometric.estimateSpectralBinning.EstimateSpectralBinning` by default.

        Returns
        -------
        :class:`~astropy.table.QTable`
            table for the radiometric estimations with wavelength grid.
        """

        self.info("Computing radiometric signals")
        estimateSpectralBinning_task = (
            find_task(
                self.payloadConfig["channel"]["spectral_binning_task"],
                radiometric.EstimateSpectralBinning,
            )
            if "spectral_binning_task" in self.payloadConfig["channel"].keys()
            else radiometric.EstimateSpectralBinning
        )
        estimateSpectralBinning = estimateSpectralBinning_task()
        if isinstance(self.payloadConfig["channel"], OrderedDict):
            table_list = [
                estimateSpectralBinning(
                    parameters=self.payloadConfig["channel"][ch]
                )
                for ch in self.payloadConfig["channel"].keys()
            ]
        else:
            table_list = [
                estimateSpectralBinning(
                    parameters=self.payloadConfig["channel"]
                )
            ]
        return vstack(table_list)

    def compute_apertures(self) -> QTable:
        """
        Estimates the photometric aperture for each spectral bin
        using :class:`~exosim.tasks.radiometric.estimateApertures.EstimateApertures` by default.

        Returns
        -------
        :class:`~astropy.table.QTable`
            table with the apertures for each channel and bin
        """
        if isinstance(self.payloadConfig["channel"], OrderedDict):
            table_list = []
            for ch in self.payloadConfig["channel"].keys():
                self.info("estimating apertures on {}".format(ch))
                description = self.payloadConfig["channel"][ch]
                estimateApertures_task = (
                    find_task(
                        description["radiometric"]["aperture_photometry"][
                            "apertures_task"
                        ],
                        radiometric.EstimateApertures,
                    )
                    if "apertures_task"
                    in description["radiometric"]["aperture_photometry"].keys()
                    else radiometric.EstimateApertures
                )
                estimateApertures = estimateApertures_task()

                with self.input.open() as f:
                    f = f["channels"]
                    self.debug("extracting wavelength grid")
                    osf = f[ch]["focal_plane"]["metadata"]["oversampling"][()]
                    focal_plane = f[ch]["focal_plane"]["data"][
                        0, osf // 2 :: osf, osf // 2 :: osf
                    ]
                    wl_grid = f[ch]["focal_plane"]["spectral"][
                        osf // 2 :: osf
                    ] * u.Unit(f[ch]["focal_plane"]["spectral_units"][()])

                    table_list += [
                        estimateApertures(
                            table=self.table[self.table["ch_name"] == ch],
                            focal_plane=focal_plane,
                            description=description["radiometric"][
                                "aperture_photometry"
                            ],
                            wl_grid=wl_grid,
                        )
                    ]
        else:
            description = self.payloadConfig["channel"]
            self.info(
                "estimating apertures on {}".format(description["value"])
            )
            estimateApertures_task = (
                find_task(
                    description["radiometric"]["aperture_photometry"][
                        "apertures_task"
                    ],
                    radiometric.EstimateApertures,
                )
                if "apertures_task"
                in description["radiometric"]["aperture_photometry"].keys()
                else radiometric.EstimateApertures
            )
            estimateApertures = estimateApertures_task()
            with self.input.open() as f:
                f = f["channels"]
                ch = list(f.keys())[0]
                self.debug("extracting wavelength grid")
                osf = f[ch]["focal_plane"]["metadata"]["oversampling"][()]
                focal_plane = f[ch]["focal_plane"]["data"][
                    0, osf // 2 :: osf, osf // 2 :: osf
                ]
                wl_grid = f[ch]["focal_plane"]["spectral"][
                    osf // 2 :: osf
                ] * u.Unit(f[ch]["focal_plane"]["spectral_units"][()])

                table_list = [
                    estimateApertures(
                        table=self.table,
                        focal_plane=focal_plane,
                        wl_grid=wl_grid,
                        description=description["radiometric"][
                            "aperture_photometry"
                        ],
                    )
                ]

        stack = vstack(table_list)
        self.table = hstack((self.table, stack))
        return hstack((self.table["ch_name", "Wavelength"], stack))

    def write(self, output_file: str = None) -> None:
        """
        It adds the radiometric table to the output.
        If the table exists already in the output file, it replaces it.

        Parameters
        ----------
        output_file: str (optional)
            output file. Default is `input`

        """
        output = (
            self.input
            if output_file is None
            else SetOutput(output_file, replace=False)
        )

        with output.use(append=True) as out:
            self.info("radiometric table stored in {}".format(output.fname))
            rad_out = out.create_group("radiometric")
            rad_out.write_table("table", self.table, replace=True)

    def compute_sub_foregrounds_signals(self) -> QTable:
        """
        It estimates the radiometric signals on the foreground sub focal planes for all the
        channels and returns a table with all the contributions.

        It uses :class:`~exosim.tasks.radiometric.computeSubFrgSignalsChannel.ComputeSubFrgSignalsChannel` by default.

        Returns
        -------
        astropy.table.QTable:
            signal table
        """

        table_list = []
        if isinstance(self.payloadConfig["channel"], OrderedDict):
            for ch in self.payloadConfig["channel"].keys():
                self.info(
                    "estimating sub-foreground radiometric signal for {}".format(
                        ch
                    )
                )
                computeFrgSignalsChannel_task = (
                    find_task(
                        self.payloadConfig["channel"][ch]["radiometric"][
                            "sub_frg_signal_task"
                        ],
                        radiometric.ComputeSubFrgSignalsChannel,
                    )
                    if "sub_frg_signal_task"
                    in self.payloadConfig["channel"][ch]["radiometric"].keys()
                    else radiometric.ComputeSubFrgSignalsChannel
                )
                computeFrgSignalsChannel = computeFrgSignalsChannel_task()
                table_list += [
                    computeFrgSignalsChannel(
                        table=self.table[self.table["ch_name"] == ch],
                        ch_name=ch,
                        input_file=self.input,
                        parameters=self.payloadConfig["channel"],
                    )
                ]
        else:
            self.info(
                "estimating sub-foreground radiometric signal for {}".format(
                    self.payloadConfig["channel"]["value"]
                )
            )
            computeFrgSignalsChannel_task = (
                find_task(
                    self.payloadConfig["channel"]["radiometric"][
                        "sub_frg_signal_task"
                    ],
                    radiometric.ComputeSubFrgSignalsChannel,
                )
                if "sub_frg_signal_task"
                in self.payloadConfig["channel"]["radiometric"].keys()
                else radiometric.ComputeSubFrgSignalsChannel
            )
            computeFrgSignalsChannel = computeFrgSignalsChannel_task()
            table_list = [
                computeFrgSignalsChannel(
                    table=self.table,
                    ch_name=self.payloadConfig["channel"]["value"],
                    input_file=self.input,
                )
            ]

        stack = vstack(table_list)
        self.table = hstack((self.table, stack))
        for k in self.table.keys():
            if hasattr(self.table[k], "filled"):
                self.table[k] = self.table[k].filled(0.0)
        ret_k = ["ch_name", "Wavelength"] + list(stack.keys())
        return self.table[ret_k], stack

    def compute_foreground_signals(self) -> QTable:
        """
        It estimates the radiometric signals on the foreground focal plane for all the
        channels and returns a table with all the contributions.

        It uses :class:`~exosim.tasks.radiometric.computeSignalsChannel.ComputeSignalsChannel` by default.

        Returns
        -------
        astropy.table.QTable:
            signal table
        """

        phot = np.array([])
        with self.input.open() as f:
            if isinstance(self.payloadConfig["channel"], OrderedDict):
                for ch in self.payloadConfig["channel"].keys():
                    self.info(
                        "estimating foreground radiometric signal for {}".format(
                            ch
                        )
                    )
                    computeSignalsChannel_task = (
                        find_task(
                            self.payloadConfig["channel"][ch]["radiometric"][
                                "signal_task"
                            ],
                            radiometric.ComputeSignalsChannel,
                        )
                        if "signal_task"
                        in self.payloadConfig["channel"][ch][
                            "radiometric"
                        ].keys()
                        else radiometric.ComputeSignalsChannel
                    )
                    computeSignalsChannel = computeSignalsChannel_task()

                    focal_plane = f["channels"][ch]["frg_focal_plane"]
                    phot_ = computeSignalsChannel(
                        table=self.table[self.table["ch_name"] == ch],
                        focal_plane=focal_plane,
                    )
                    phot = np.concatenate((phot, phot_))
            else:
                self.info(
                    "estimating foreground radiometric signal for {}".format(
                        self.payloadConfig["channel"]["value"]
                    )
                )
                computeSignalsChannel_task = (
                    find_task(
                        self.payloadConfig["channel"]["radiometric"][
                            "signal_task"
                        ],
                        radiometric.ComputeSignalsChannel,
                    )
                    if "signal_task"
                    in self.payloadConfig["channel"]["radiometric"].keys()
                    else radiometric.ComputeSignalsChannel
                )
                computeSignalsChannel = computeSignalsChannel_task()

                ch = list(f["channels"].keys())[0]
                focal_plane = f["channels"][ch]["frg_focal_plane"]
                phot_ = computeSignalsChannel(
                    table=self.table, focal_plane=focal_plane
                )
                phot = np.concatenate((phot, phot_))

        self.table["foreground_signal_in_aperture"] = phot
        return self.table[
            "ch_name", "Wavelength", "foreground_signal_in_aperture"
        ]

    def compute_source_signals(self) -> QTable:
        """
        It estimates the radiometric signals on the source focal plane for all the
        channels and returns a table with all the contributions.

        It uses :class:`~exosim.tasks.radiometric.computeSignalsChannel.ComputeSignalsChannel` by default.

        Returns
        -------
        astropy.table.QTable:
            signal table
        """

        phot = np.array([])
        with self.input.open() as f:
            if isinstance(self.payloadConfig["channel"], OrderedDict):
                for ch in self.payloadConfig["channel"].keys():
                    self.info(
                        "estimating source radiometric signal for {}".format(
                            ch
                        )
                    )
                    computeSignalsChannel_task = (
                        find_task(
                            self.payloadConfig["channel"][ch]["radiometric"][
                                "signal_task"
                            ],
                            radiometric.ComputeSignalsChannel,
                        )
                        if "signal_task"
                        in self.payloadConfig["channel"][ch][
                            "radiometric"
                        ].keys()
                        else radiometric.ComputeSignalsChannel
                    )
                    computeSignalsChannel = computeSignalsChannel_task()

                    focal_plane = f["channels"][ch]["focal_plane"]
                    phot_ = computeSignalsChannel(
                        table=self.table[self.table["ch_name"] == ch],
                        focal_plane=focal_plane,
                    )
                    phot = np.concatenate((phot, phot_))

            else:
                self.info(
                    "estimating source radiometric signal for {}".format(
                        self.payloadConfig["channel"]["value"]
                    )
                )
                computeSignalsChannel_task = (
                    find_task(
                        self.payloadConfig["channel"]["radiometric"][
                            "signal_task"
                        ],
                        radiometric.ComputeSignalsChannel,
                    )
                    if "signal_task"
                    in self.payloadConfig["channel"]["radiometric"].keys()
                    else radiometric.ComputeSignalsChannel
                )
                computeSignalsChannel = computeSignalsChannel_task()

                ch = list(f["channels"].keys())[0]
                focal_plane = f["channels"][ch]["focal_plane"]
                phot_ = computeSignalsChannel(
                    table=self.table, focal_plane=focal_plane
                )
                phot = np.concatenate((phot, phot_))

        self.table["source_signal_in_aperture"] = phot
        return self.table["ch_name", "Wavelength", "source_signal_in_aperture"]

    def compute_saturation(self) -> QTable:
        """
        It computes and adds the saturation time to the radiometric table

        Returns
        -------
        astropy.table.QTable:
            saturation table
        """
        saturations, integration_time, max_signal_in_bin, min_signal_in_bin = (
            [],
            [],
            [],
            [],
        )

        saturationChannel = radiometric.SaturationChannel()
        if isinstance(self.payloadConfig["channel"], OrderedDict):
            for ch in self.payloadConfig["channel"].keys():
                sat, t_int, max_, min_ = saturationChannel(
                    table=self.table,
                    description=self.payloadConfig["channel"][ch],
                    input_file=self.input,
                )
                saturations += sat
                integration_time += t_int
                max_signal_in_bin += max_
                min_signal_in_bin += min_
        else:
            sat, t_int, max_, min_ = saturationChannel(
                table=self.table,
                description=self.payloadConfig["channel"],
                input_file=self.input,
            )
            saturations += sat
            integration_time += t_int
            max_signal_in_bin += max_
            min_signal_in_bin += min_

        self.table["saturation_time"] = saturations
        self.table["integration_time"] = integration_time
        self.table["max_signal_in_bin"] = max_signal_in_bin
        self.table["min_signal_in_bin"] = min_signal_in_bin

        return self.table[
            "ch_name",
            "Wavelength",
            "saturation_time",
            "integration_time",
            "max_signal_in_bin",
            "min_signal_in_bin",
        ]

    def compute_multiaccum(self) -> QTable:
        """
        It estimates the multiaccum gain factors using :class:`~exosim.tasks.radiometric.multiaccum.Multiaccum`. The

        Returns
        -------
        astropy.table.QTable:
           multiaccum factors
        """

        read_gain, shot_gain = [], []

        multiaccum = radiometric.Multiaccum()
        if isinstance(self.payloadConfig["channel"], OrderedDict):
            for ch in self.payloadConfig["channel"].keys():
                if (
                    "multiaccum"
                    in self.payloadConfig["channel"][ch]["radiometric"].keys()
                ):
                    read, shot = multiaccum(
                        parameters=self.payloadConfig["channel"][ch][
                            "radiometric"
                        ]["multiaccum"]
                    )
                    read_gain += [read] * len(
                        self.table[self.table["ch_name"] == ch]
                    )
                    shot_gain += [shot] * len(
                        self.table[self.table["ch_name"] == ch]
                    )
                else:
                    read_gain += [1] * len(
                        self.table[self.table["ch_name"] == ch]
                    )
                    shot_gain += [1] * len(
                        self.table[self.table["ch_name"] == ch]
                    )

        else:
            if (
                "multiaccum"
                in self.payloadConfig["channel"]["radiometric"].keys()
            ):
                read, shot = multiaccum(
                    parameters=self.payloadConfig["channel"]["radiometric"][
                        "multiaccum"
                    ]
                )
                read_gain += [read] * len(self.table)
                shot_gain += [shot] * len(self.table)
            else:
                read_gain += [1] * len(self.table)
                shot_gain += [1] * len(self.table)

        self.table["multiaccum_read_gain"] = read_gain
        self.table["multiaccum_shot_gain"] = shot_gain
        return self.table[
            "ch_name",
            "Wavelength",
            "multiaccum_read_gain",
            "multiaccum_shot_gain",
        ]

    def compute_photon_noise(self) -> QTable:
        """
        It computes and adds the photon noise to the radiometric table using :class:`~exosim.tasks.radiometric.computePhotonNoise.ComputePhotonNoise`.

        Returns
        -------
        astropy.table.QTable:
           photon noise
        """

        computePhotonNoise = radiometric.ComputePhotonNoise()
        signals = [k for k in self.table.keys() if "_signal_in_aperture" in k]
        for sig in signals:
            phot_noise = []
            for i in range(len(self.table["ch_name"])):
                ch_description = (
                    self.payloadConfig["channel"][self.table["ch_name"][i]]
                    if isinstance(self.payloadConfig["channel"], OrderedDict)
                    else self.payloadConfig["channel"]
                )
                phot_noise += [
                    computePhotonNoise(
                        signal=self.table[sig][i],
                        description=ch_description,
                        multiaccum_gain=self.table["multiaccum_shot_gain"][i],
                    )
                ]
            name = sig.replace("_signal_in_aperture", "")

            phot_noise = (
                np.array([p.value for p in phot_noise]) * phot_noise[0].unit
            )
            self.table["{}_photon_noise".format(name)] = phot_noise

        return self.table[
            "ch_name",
            "Wavelength",
            "source_photon_noise",
            "foreground_photon_noise",
        ]

    def update_total_noise(self) -> QTable:
        """Updates the total noise column in the radiometric table."""
        computeTotalNoise = radiometric.ComputeTotalNoise()
        total_noise = computeTotalNoise(table=self.table)
        self.table["total_noise"] = total_noise
        return self.table["ch_name", "Wavelength", "total_noise"]



import astropy.units as u
import numpy as np

import exosim.output as output
import exosim.utils as utils
from exosim.tasks.task import Task
from exosim.utils import RunConfig
from exosim.utils.checks import check_units


class EstimatePointingJitter(Task):
    """
    Produces the telescope pointing jitter expressed as deg on the line of sight.

    Returns
    --------
    :class:`~astropy.units.Quantity`,
        pointing jitter in the spatial and spectral direction expressed in units of :math:`deg`.
    :class:`~astropy.units.Quantity`
        pointing jitter in the spatial and spectral direction expressed in units of :math:`deg`.
    :class:`~astropy.units.Quantity`
        pointing jitter in the spatial and spectral direction expressed in units of :math:`deg`.

    Raises
    ------
    TypeError:
        if the output is not a :class:`~astropy.units.Quantity` class

    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        """
        self.add_task_param("parameters", "main parameters dict")
        self.add_task_param("output_file", "output file", None)

    def execute(self):
        self.info("estimating pointing jitter")
        main_parameters = self.get_task_param("parameters")

        jitter_spa, jitter_spe, jitter_time = self.model(main_parameters)

        # checking output
        if not isinstance(jitter_spa, u.Quantity):
            self.error("wrong output format")
            raise TypeError("wrong output format")
        if not isinstance(jitter_spe, u.Quantity):
            self.error("wrong output format")
            raise TypeError("wrong output format")
        if not isinstance(jitter_time, u.Quantity):
            self.error("wrong output format")
            raise TypeError("wrong output format")
        # converting into dec
        jitter_spa = utils.check_units(jitter_spa, u.deg)
        jitter_spe = utils.check_units(jitter_spe, u.deg)
        jitter_time = utils.check_units(jitter_time, u.s)

        self.debug("pointing jitter: {} {}".format(jitter_spa, jitter_spe))
        output_file = self.get_task_param("output_file")
        if output_file:
            if issubclass(output_file.__class__, output.Output):
                store_dict = {
                    "jitter_spa": jitter_spa,
                    "jitter_spe": jitter_spe,
                    "jitter_time": jitter_time,
                }
                output_file.store_dictionary(store_dict, "pointing_jitter")

        self.set_output([jitter_spa, jitter_spe, jitter_time])

    def model(self, parameters):
        """
        This default model builds the pointing jitter using random values.
        Starting from the spectral and spatial standard deviations, expressed as angles,
        it computes the pointing position as normally distributed around zero in the two direction.
        The input dictionary, under the `jitter` keyword, must contain `spatial` and `spectral` keyword for the standard deviation.
        The time grid is built using the `high_frequencies_resolution` under the `time_grid` keyword.


        Parameters
        __________
        parameters: dict
            dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`

        Returns
        --------
        :class:`~astropy.units.Quantity`
            pointing jitter in the spatial direction expressed in units of :math:`deg`.
         :class:`~astropy.units.Quantity`)
            pointing jitter in the spectral direction expressed in units of :math:`deg`.
         :class:`~astropy.units.Quantity`)
            pointing jitter timeline expressed in units of :math:`s`.
        """
        jitter_time_step = check_units(
            parameters["jitter"]["frequency_resolution"], "s"
        )

        jitter_time = utils.grids.time_grid(
            parameters["time_grid"]["start_time"],
            parameters["time_grid"]["end_time"]
            + 1 * parameters["time_grid"]["end_time"].unit,
            jitter_time_step,
        )

        jitter_spa = (
            RunConfig.random_generator.normal(
                0, parameters["jitter"]["spatial"].value, jitter_time.size
            )
            * parameters["jitter"]["spatial"].unit
        )
        jitter_spe = (
            RunConfig.random_generator.normal(
                0, parameters["jitter"]["spectral"].value, jitter_time.size
            )
            * parameters["jitter"]["spectral"].unit
        )

        return jitter_spa, jitter_spe, jitter_time



import astropy.units as u
from astropy.modeling.physical_models import BlackBody

import exosim.models.signal as signal
import exosim.utils.checks as checks
from exosim.tasks.task import Task


class EstimateZodi(Task):
    """
    It estimate the zodiacal radiance in the target direction for a specific wavelength range

    Returns
    -------
    :class:`~exosim.models.signal.Radiance`
        zodiacal radiance

    Examples
    --------
    >>> estimateZodi = EstimateZodi()
    >>> wavelength = np.logspace(np.log10(0.45), np.log10(2.2), 6000) * u.um
    >>> zodi = estimateZodi(wavelength=wavelength, zodiacal_factor=1)

    or, given the pointing direction

    >>> zodi = self.estimateZodi(wavelength=wavelength,  coordinates=(90 * u.deg, -66 * u.deg))

    """

    def __init__(self):
        """
        Parameters
        -----------
        wavelength: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity`
                wavelength grid. If no units are attached is considered as expressed in `um`
        zodiacal_factor: float
            zodiacal model multiplicative factor
        coordinates: (float, float)
            pointing coordinates as (ra, dec)
        zodi_map: str
            map file containing the zodiacal coefficients per sky positions. A default map is inclluded in ExoSim,
            that contains the coefficients fitted over Kelsall et al. 1998 model.
        """
        self.add_task_param("wavelength", "wavelength range to investigate")
        self.add_task_param(
            "zodiacal_factor", "zodiacal multiplicative factor", None
        )
        self.add_task_param("coordinates", "pointing direction", None)
        self.add_task_param(
            "zodi_map",
            " map file containing the zodiacal coefficients per sky positions",
            None,
        )

    def execute(self):
        self.info("estimating zodiacal foreground")
        wl = self.get_task_param("wavelength")
        zodiacal_factor = self.get_task_param("zodiacal_factor")
        coordinates = self.get_task_param("coordinates")
        zodi_map = self.get_task_param("zodi_map")

        wl = checks.check_units(wl, u.um, self)

        if coordinates:
            zodiacal_factor = self.zodiacal_fit_direction(
                coordinates, zodi_map
            )

        rad = self.model(zodiacal_factor, wl)
        self.debug("zodical radiance: {}".format(rad.data))

        self.set_output(rad)

    def model(self, a, wl):
        r"""
        The used zodiacal model is based on the zodiacal light model presented in Glasse et al. 2010

        .. math::

            I_{zodi}(\lambda) = a \left( 3.5 \cdot 10^{-14} BB(\lambda, 5500 \, K) + 3.52 \cdot 10^{-8} BB(\lambda, 270 \, K) \right)

        where :math:`BB(\lambda, T)` is the Planck black body law and :math:`a` is the fitted coefficient.

        Parameters
        ----------
        a: float
            zodiacal multiplicative factor. Default is 0.
        wl: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity`
            wavelength grid. If no units are attached is considered as expressed in `um`

        Returns
        -------
        :class:`~exosim.models.signal.Radiance`
            zodiacal radiance

        """
        if not a:
            a = 0
        units = u.W / (u.m**2 * u.um * u.sr)
        bb_1 = BlackBody(5500.0 * u.K)
        bb_1 = bb_1(wl).to(u.W / u.m**2 / u.sr / u.um, u.spectral_density(wl))
        bb_2 = BlackBody(270.0 * u.K)
        bb_2 = bb_2(wl).to(u.W / u.m**2 / u.sr / u.um, u.spectral_density(wl))

        zodi_emission = a * (3.5e-14 * bb_1 + 3.58e-8 * bb_2).to(units)
        return signal.Radiance(wl, zodi_emission)

    def zodiacal_fit_direction(self, coordinates, zodi_map=None):
        r"""
        In this case the :math:`A` coefficient is selected by a precompiled grid.
        The grid has been estimated by fitting our model with Kelsall et al. (1998) data.

        A custom map can be provided, to replace the default one, as long as it matches the format.

        Parameters
        ----------
        coordinates: (float, float)
            pointing coordinates as (ra, dec)
        zodi_map: str (optional)
            map file containing the zodiacal coefficients per sky positions. A default map is inclluded in ExoSim,
            that contains the coefficients fitted over Kelsall et al. 1998 model.

        Returns
        -------
        float
            zodiacal factor for Glasse et al. 2010 zodiacal model

        """
        import os
        from pathlib import Path

        import numpy as np
        from astropy.io.misc.hdf5 import read_table_hdf5

        ra_input = coordinates[0]
        dec_input = coordinates[1]

        if zodi_map:
            zodi_map_file = zodi_map
        else:
            dir_path = Path(os.path.dirname(os.path.realpath(__file__)))
            i = 0
            while (
                "data"
                not in [d.stem for d in Path(dir_path).iterdir() if d.is_dir()]
                or i > 10
            ):
                dir_path = dir_path.parent
                i += 1
            if i > 10:
                self.error("default zodi map file not found")
                raise OSError("default zodi map file not found")
            data_path = os.path.join(dir_path.absolute().as_posix(), "data")
            zodi_map_file = os.path.join(data_path, "Zodi_map.hdf5")
            self.debug("map data:{}".format(zodi_map_file))

        try:
            zodi_table = read_table_hdf5(zodi_map_file)
            self.debug(zodi_table)
            # here we find the minimun distance from the desired point
            distance = (zodi_table["ra_icrs"] * u.deg - ra_input) ** 2 + (
                zodi_table["dec_icrs"] * u.deg - dec_input
            ) ** 2
            idx = np.argmin(distance)
            # TODO implement an interpolator
            self.debug("selected line {}".format(idx))
            self.debug(zodi_table[idx])
            return zodi_table["zodi_coeff"][idx]

        except OSError:
            self.error("Zodi map file not found")
            raise OSError("Zodi map file not found")



import os
import tempfile
import uuid
from pathlib import Path
from typing import Union

import h5py
import numpy as np

import exosim.log as log
from exosim.utils import RunConfig
from exosim.utils.types import HDF5OutputType


class CachedData(log.Logger):
    """
    This class caches data cube into an h5 file. The cube data are chunked toward the first axis.
    In this class are also defined a set of operation to operate on the dataset using the chinks system.

    Attributes
    ----------
    axis0: int
        first axis size
    axis1: int
        second axis size
    axis2: int
        third axis size
    output: str or :class:`~exosim.output.HDF5Output`
        name of the file used for caching.
    dataset_name: str
        name used to store the dataset into the h5 file.
    output: :class:`h5py.File` or  :class:`~exosim.output.hdf5.hdf5.HDF5Output` or :class:`~exosim.output.hdf5.hdf5.HDF5OutputGroup`
        h5py open file used for caching
    dataset_path: str
         path where is stored the dataset inside the output file.
    chunked_dataset: :class:`h5py.Dataset`
        h5py dataset used to store the data

    Notes
    -----
    The cached data may be stored in a temporary file. To delete temporary files we included a garbage collector.
    Please, remember to delete the class when done as in the following example

    >>> myClass = CachedData(1,1,1)
    >>> del myClass
    """

    def __init__(
        self,
        axis0: int,
        axis1: int,
        axis2: int,
        output: Union[str, "HDF5Output", "HDF5OutputGroup"] = None,  # type: ignore # noqa: F821
        output_path: str = None,
        dataset_name: str = None,
        dtype: np.dtype = np.float64,
    ) -> None:
        """
        Parameters
        ----------
        axis0: int
            first axis size
        axis1: int
            second axis size
        axis2: int
            third axis size
        output: str or :class:`~exosim.output.hdf5.hdf5.HDF5Output` or :class:`~exosim.output.hdf5.hdf5.HDF5OutputGroup`
            file name to use for caching. If `None` a temporary file will be generated. Default is `None`.
        output_path: str (optional)
            path where to store the dataset inside the output file. Default is `None`.
        dataset_name: str
            name to use to store the dataset into the h5 file. If `None` a random name will be generated. Default is `None`.
        """
        self.set_log_name()

        self.axis0 = axis0
        self.axis1 = axis1
        self.axis2 = axis2
        self.output = None

        # we want to get a name for the dataset to store it in the h5 file
        if dataset_name is None:
            dataset_name = str(uuid.uuid4())[:8]
        self.dataset_name = dataset_name
        self.debug("data will be stored as: {}".format(self.dataset_name))

        # if a file name is given, we use that as output, otherwise we use a temporary file
        self.tmp = False
        if output:
            # if is a string or an already existing temporary file, we open it as an h5py
            if isinstance(output, str):
                self.output = h5py.File(output, "a", rdcc_w0=1)
                self.fname = output
            elif isinstance(output, tempfile._TemporaryFileWrapper):
                self.output = h5py.File(output, "a", rdcc_w0=1)
                self.fname = output
                self.tmp = True
            # if it is an Output class, we use it
            else:
                from exosim.output import HDF5Output, HDF5OutputGroup

                if isinstance(output, HDF5Output):
                    if not output._cache:
                        self.error("output file not set for caching")
                        raise OSError("output file not set for caching")
                    self.fname = output.filename
                    self.output = output.fd
                elif isinstance(output, HDF5OutputGroup):
                    if not output._cache:
                        self.error("output file not set for caching")
                        raise OSError("output file not set for caching")
                    self.fname = output.filename
                    self.output = output._entry

                else:
                    self.error("unsupported output format")
                    raise OSError("unsupported output format")

            self.debug("data stored in: {}".format(self.fname))
        else:
            path = Path(os.path.join(os.getcwd(), "tmp"))
            tempfile.tempdir = path
            tempfile.tempdir.mkdir(parents=True, exist_ok=True)
            self.fname = tempfile.NamedTemporaryFile(
                suffix=".h5", delete=False
            )
            self.output = h5py.File(self.fname, "a", overrite=True, rdcc_w0=1)
            self.tmp = True
            self.debug("temporary file created: {}".format(self.fname.name))

        # we define the address of the dataset inside the file
        self.output_path = output_path
        if self.output_path:
            self.dataset_path = os.path.join(output_path, self.dataset_name)
        else:
            self.dataset_path = self.dataset_name + "/data"

        # chunk size fixed to 2Mb
        mem_size = RunConfig.chunk_size * 1e6  # convert to Mbs
        image_size = axis1 * axis2 * np.dtype(dtype).itemsize
        axis0_chunk = (
            int(mem_size // image_size)
            if int(mem_size // image_size) < axis0
            else axis0
        )
        # we create an empty dataset (full of zeros) of a given shape, chunked of the first axis
        self.chunked_dataset = self.output.create_dataset(
            self.dataset_path,
            shape=(axis0, axis1, axis2),
            chunks=(axis0_chunk, axis1, axis2),
            dtype=dtype,
            compression=None,
        )

    def rename_dataset(self, new_name: str) -> None:
        """It renames the dataset in the HDF5 file.

        Parameters
        -----------
        new_name: str
           new name for the dataset
        """
        self.output[new_name] = self.output[self.dataset_name]
        del self.output[self.dataset_name]
        self.dataset_name = new_name

    # to perform any operation, we iterate over the chunks built on the first axis
    # def _get_val_in_slice(self, other, i):
    #     """Returs the values contained in chunk built on the first axis."""
    #     if hasattr(other, 'dataset'):
    #         return other.dataset[i, :, :]
    #     elif hasattr(other, 'shape'):
    #         return other[i, :, :]
    #     else:
    #         return np.ones(self.dataset[i, :, :].shape) * other
    #
    # def _create_new_instance(self):
    #     return CachedData(*self.dataset.shape, output=self.fname,
    #                       output_path=self.output_path)
    #
    # @property
    # def dataset(self):
    #     """h5py dataset used to store the data"""
    #     return self._dataset
    #
    # @dataset.setter
    # def dataset(self, other):
    #     """ This setter implements the slicing/caching system"""
    #     self._dataset = other
    #     self.output.flush()
    #
    # @dataset.deleter
    # def dataset(self):
    #     """It empties the dataset"""
    #     del self._dataset

    def __del__(self):
        """Garbage collector"""
        # close the file

        # try to flush and close the output file
        if self.output and isinstance(self.output, h5py.File):
            if hasattr(self.output, "flush"):
                try:
                    self.output.flush()
                except ValueError:
                    pass
            if hasattr(self.output, "close"):
                self.output.close()

        if self.tmp:
            # remove temp file if exist
            try:
                os.remove(self.fname)
                self.debug("file deleted: {}".format(self.fname))
            except TypeError:
                os.remove(self.fname.name)
                self.debug(
                    "temporary file deleted: {}".format(self.fname.name)
                )
            except FileNotFoundError:
                pass

            # check if the file has been correctly removed
            try:
                if Path(self.fname).is_file():
                    self.warning("file not deleted: {}".format(self.fname))
            except TypeError:
                if Path(self.fname.name).is_file():
                    self.warning(
                        "file not deleted: {}".format(self.fname.name)
                    )

            # if temp dir is empty, delete it
            try:
                if not any(tempfile.tempdir.iterdir()):
                    os.rmdir(tempfile.tempdir)
                    self.debug(
                        "temporary dir deleted: {}".format(tempfile.tempdir)
                    )
            except FileNotFoundError:
                pass
            except AttributeError:
                pass



import astropy.units as u
from astropy.table import QTable

from exosim.tasks.task import Task


class LoadWavelengthSolution(Task):
    """
    It loads the wavelength solution expressed as a table with wavelength position on spectral and spatial directcions

    Returns
    --------
    :class:`~astropy.table.Qtable`
        wavelength solution table

    Raises
    -------
    TypeError:
        if the output is not a :class:`~astropy.table.QTable` class

    UnitsError
        if the output has not the right units (micron) in every column

    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            channel parameter dictionary. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        """

        self.add_task_param("parameters", "channel parameters dict")

    def execute(self):
        parameters = self.get_task_param("parameters")

        wl_solution = self.model(parameters)

        if not isinstance(wl_solution, QTable):
            self.error("wrong output format")
            raise TypeError("wrong output format")
        for k in ["wavelength", "spectral", "spatial"]:
            if k not in wl_solution.keys():
                self.error("missing {} keyword".format(k))
                raise KeyError("missing {} keyword".format(k))
            elif wl_solution[k].unit != u.um:
                self.error("wrong units: {} units are not micron".format(k))
                raise u.UnitsError(
                    "wrong units: {} units are not micron".format(k)
                )

        self.set_output(wl_solution)

    def model(self, parameters):
        """
        It loads the Wavelength, X and Y columns from a file and it stores them into a class:`~astropy.table.Qtable` class

        Parameters
        ----------
        parameters: dict
            channel parameter dictionary. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`


        Returns
        --------
        :class:`~astropy.table.Qtable`
            wavelength solution table

        """
        wl_solution = QTable()
        wl_solution["wavelength"] = parameters["wl_solution"]["data"][
            "Wavelength"
        ]
        wl_solution["spectral"] = parameters["wl_solution"]["data"]["x"]
        wl_solution["spatial"] = parameters["wl_solution"]["data"]["y"]
        return wl_solution



import astropy.units as u
from joblib import Parallel, delayed

from exosim.tasks.instrument import ComputeSolidAngle
from exosim.tasks.task import Task
from exosim.utils import RunConfig


class PropagateForegrounds(Task):
    """
    it propagates the foreground though the channel.

    Returns
    -------
    `~collections.OrderedDict`
        dictionary of :class:`~exosim.models.signal.Radiance` and :class:`~exosim.models.signal.Dimensionless`,
        represeting the radiance and efficiency of the path.
    """

    def __init__(self):
        """
        Parameters
        __________
        light_path: `~collections.OrderedDict`
            dictionary of :class:`~exosim.models.signal.Radiance` and :class:`~exosim.models.signal.Dimensionless`,
            represeting the radiance and efficiency of the path.
        responsivity:  :class:`~exosim.models.signal.Signal`
            channel responsivity
        parameters: dict
            channel parameter dictionary. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        """

        self.add_task_param("light_path", "sources dictionary")
        self.add_task_param("responsivity", "channel responsivity")
        self.add_task_param("parameters", "channel parameters dict")

    def execute(self):
        light_path = self.get_task_param("light_path")
        parameters = self.get_task_param("parameters")
        responsivity = self.get_task_param("responsivity")

        # for rad in [k for k in light_path.keys() if 'radiance' in k]:
        #     computeSolidAngle = ComputeSolidAngle()
        #     solid_angle = computeSolidAngle(parameters=parameters,
        #                                     other_parameters=light_path[
        #                                         rad].metadata)
        #     light_path[rad] *= solid_angle * responsivity
        #     light_path[rad].to(u.ct / u.s / u.um)
        computeSolidAngle = ComputeSolidAngle()
        Parallel(n_jobs=RunConfig.n_job, require="sharedmem")(
            delayed(self._computation_model)(
                rad,
                computeSolidAngle,
                parameters,
                light_path[rad].metadata,
                responsivity,
                light_path,
            )
            for rad in [k for k in light_path.keys() if "radiance" in k]
        )

        self.set_output(light_path)

    def _computation_model(
        self,
        rad,
        computeSolidAngle,
        parameters,
        other_parameters,
        responsivity,
        light_path,
    ):
        solid_angle = computeSolidAngle(
            parameters=parameters, other_parameters=other_parameters
        )
        light_path[rad] *= solid_angle * responsivity
        light_path[rad].to(u.ct / u.s / u.um)



from photutils.aperture import (
    EllipticalAperture,
    RectangularAperture,
    aperture_photometry,
)

from exosim.tasks.task import Task


class AperturePhotometry(Task):
    """
    It performs the aperture photometry using :func:`photutils.aperture.aperture_photometry`.

    The details of the aperture strongly depends on the configurations set by the user.

    """

    def __init__(self):
        """
        Parameters
        --------------
        table: :class:`astropy.table.QTable`
            wavelength table with bin edge
        focal_plane:
            focal plane
        """
        self.add_task_param(
            "table", "wavelength table with bin edges and aperture sizes"
        )
        self.add_task_param("focal_plane", "focal plane")

    def execute(self):
        self.debug("performing aperture photometry")
        table = self.get_task_param("table")
        focal_plane = self.get_task_param("focal_plane")

        aperture_shapes = {
            "rectangular": RectangularAperture,
            "elliptical": EllipticalAperture,
        }

        photometry = []
        for i in range(table["Wavelength"].size):
            # wavelength position estimated from the focal plane wavelength solution
            center_spectral = table["spectral_center"][i]
            spectral_size = table["spectral_size"][i]
            spatial_size = table["spatial_size"][i]
            center_spatial = table["spatial_center"][i]
            shape = table["aperture_shape"][i]

            aper = aperture_shapes[shape](
                (center_spectral, center_spatial),
                spectral_size,
                spatial_size,
            )

            phot = aperture_photometry(focal_plane, aper)
            photometry.append(phot["aperture_sum"].data[0])

        self.set_output(photometry)



import re
from collections import OrderedDict

import astropy.units as u
import numpy as np

import exosim.models.signal as signal
import exosim.output as output
from exosim.tasks.parse import ParseOpticalElement
from exosim.tasks.task import Task


class ParsePath(Task):
    """
    Given the optical path descrition, it parses the optical elements and return an ordered dictionary.


    Returns
    -------
    dict
        dictionary of :class:`~exosim.models.signal.Radiance` and :class:`~exosim.models.signal.Dimensionless`,
        represeting the radiance and efficiency of the path.

    Note
    ------
    The user can force the parser to isolate contribution by addind to the description dictionary the key 'isolate' set to ``True``.

    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            dictionary contained the optical element parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wavelength: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity`
            wavelength grid. If no units are attached is considered as expressed in `um`.
        time: :class:`~astropy.units.Quantity`
            time grid.
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        group_name: str (optional)
            group name in output
        light_path: `~collections.OrderedDict` (optional)
            dictionary of contributes
        """
        self.add_task_param("parameters", "optical element parameters dict")
        self.add_task_param("wavelength", "wavelength grid")
        self.add_task_param("time", "time grid")
        self.add_task_param("output", "output file", None)
        self.add_task_param(
            "group_name", "group name in output", "contributes"
        )
        self.add_task_param("light_path", "previous light path", None)

        self.wl, self.tt, self.out, self.radiance_key = None, None, None, None

    def execute(self):
        parameters = self.get_task_param("parameters")
        self.info("parsing optical path")

        self.wl = self.get_task_param("wavelength")
        self.tt = self.get_task_param("time")

        # prepare the output radiance key
        self.radiance_key = None
        # if a previous light path is provided
        self.out = {}
        light_path = self.get_task_param("light_path")
        if light_path:
            for key in light_path.keys():
                self.out[key] = light_path[key]
            self.radiance_key = self._find_last_radiance_key(light_path.keys())
        else:
            efficiency = signal.Dimensionless(
                spectral=self.wl,
                data=np.ones((self.tt.size, 1, self.wl.size)),
                time=self.tt,
            )
            # prepare output dict
            self.out["efficiency"] = efficiency

        output_file = self.get_task_param("output")
        group_name = self.get_task_param("group_name")
        if output_file:
            if issubclass(output_file.__class__, output.Output):
                output_file = output_file.create_group(group_name)

        # if a list of optical elements are provided we parse all of them and we sum them up
        if isinstance(parameters["opticalElement"], OrderedDict):
            for element in parameters["opticalElement"].keys():
                self._add_new_element(
                    parameters["opticalElement"][element], output_file
                )

        # else, we load the only source available
        else:
            self._add_new_element(parameters["opticalElement"], output_file)

        # output preparation
        if output_file:
            if issubclass(output_file.__class__, output.Output):
                if "final contributes" in output_file._entry.keys():
                    del output_file._entry["final contributes"]
                og = output_file.create_group("final contributes")
                for key in self.out:
                    self.out[key].write(og, key)
        self.set_output(self.out)

    def _add_new_element(self, parameters, output_file):
        """
        This method select the appropriate parser for the optical element.
        It updates the output dictionary.

        Parameters
        ----------
        parameters: dict
            parameters dict
        output_file: :class:`~exosim.output.output.Output`
            output file
        """
        # if slit, it splits the output
        if (
            "type" in parameters.keys()
            and parameters["type"].lower() == "slit"
        ):
            self._parse_slit(parameters)
        # if solid angle
        elif "solid_angle" in parameters.keys():
            self._parse_angle(parameters, output_file)
        # if isolate, store it in its own way
        elif "isolate" in parameters.keys() and parameters["isolate"]:
            # parse the optical element to get radiance and efficiency
            parsed_out = self._parse_optical_element(parameters, output_file)
            # updates efficiency
            self._update_efficiency(parsed_out)
            # check if isolate
            # updates the radiance key
            self._update_radiance_key(parameters["value"])
            # start a new radiance output
            self.out[self.radiance_key] = parsed_out["radiance"]
        # otherwise it updates it
        else:
            # if it is the first element, initialise the output dict
            if self.radiance_key is None:
                self._parse_new_radiance()
            if not self.radiance_key.split("_")[-1].isdigit():
                self._parse_new_radiance()
            if "slit_width" in self.out[self.radiance_key].metadata.keys():
                self._parse_new_radiance()
            # parse the optical element to get radiance and efficiency
            parsed_out = self._parse_optical_element(parameters, output_file)
            # updates efficiency
            self._update_efficiency(parsed_out)
            # increase current radiance
            self.out[self.radiance_key] += parsed_out["radiance"]

    def _parse_slit(self, parameters):
        """
        This method parses the slit. If a slit is found the output dict is split into two different radiances.
        It updates the output dictionary.

        Parameters
        ----------
        parameters: dict
            parameters dict
        """

        self.debug("parsing: {}".format(parameters["value"]))
        # it appends the slith width to the metadata of all the previous radiances
        for rad in [x for x in self.out.keys() if "radiance" in x]:
            self.out[rad].metadata["slit_width"] = parameters["width"]

    def _parse_angle(self, parameters, output_file):
        """
        This method parses the solid angle. If a solid angle is found the output dict is split into two different radiances.
        It updates the output dictionary.

        Parameters
        ----------
        parameters: dict
            parameters dict
        """

        self.debug("solid angle found")

        # start a new radiance output
        parseOpticalElement = ParseOpticalElement()
        parsed_out = parseOpticalElement(
            parameters=parameters,
            wavelength=self.wl,
            time=self.tt,
            output=output_file,
        )

        # updates efficiency
        self._update_efficiency(parsed_out)
        # updates the radiance key
        if "isolate" in parameters.keys() and parameters["isolate"]:
            self._update_radiance_key(parameters["value"])
        else:
            self._update_radiance_key()

        # adds the new radiance to the data
        self.out[self.radiance_key] = parsed_out["radiance"]
        # it appen the solid angle to the metadata of the previous radiance
        self.out[self.radiance_key].metadata["solid_angle"] = parameters[
            "solid_angle"
        ]

    def _parse_optical_element(self, parameters, output_file):
        if parameters["value"].lower() in ["zodiacal", "zodi"]:
            import sys  # avoid circular import

            if "ParseZodi" not in sys.modules:
                from exosim.tasks.parse.parseZodi import ParseZodi

            parseZodi = ParseZodi()
            parsed_out = parseZodi(
                parameters=parameters,
                wavelength=self.wl,
                time=self.tt,
                output=output_file,
            )
        else:
            parseOpticalElement = ParseOpticalElement()
            parsed_out = parseOpticalElement(
                parameters=parameters,
                wavelength=self.wl,
                time=self.tt,
                output=output_file,
            )
        return parsed_out

    @staticmethod
    def _find_last_radiance_key(keys_list):
        """Finds the last radiance key in the output dictionary"""
        rad_list = [rad for rad in keys_list if "radiance" in rad]
        rad_list.sort()
        return rad_list[-1]

    @property
    def radiance_keys_list(self):
        dat_re = re.compile(r"\d+")
        rad_list = [
            int(dat_re.search(rad).group())
            for rad in self.out.keys()
            if dat_re.search(rad)
        ]
        return rad_list

    def _update_radiance_key(self, name=None):
        """
        This method updates the output dictionary key for the radiance.
        """

        if self.radiance_keys_list:
            self.radiance_keys_list.sort()
            i = int(self.radiance_keys_list[-1]) + 1
        else:
            i = 0
        if name:
            self.radiance_key = "radiance_{}_{}".format(i, name)
        else:
            self.radiance_key = "radiance_{}".format(i)

        self.debug(
            "output radiance key updated to {}".format(self.radiance_key)
        )

    def _update_efficiency(self, parsed_out):
        """
        It updates the path dictionary by multiplying each efficiency and radiance by the new efficiency

        Parameters
        ----------
        parsed_out: dict
            dictionary of :class:`~exosim.models.signal.Radiance` and :class:`~exosim.models.signal.Dimensionless`,
            represeting the radiance and efficiency of the optical element.


        """
        for k in self.out.keys():
            self.out[k] *= parsed_out["efficiency"].data

    def _parse_new_radiance(self):
        self._update_radiance_key()
        radiance = signal.Radiance(
            spectral=self.wl,
            data=np.zeros((self.tt.size, 1, self.wl.size))
            * u.W
            / u.m**2
            / u.um
            / u.sr,
            time=self.tt,
        )
        # prepare output dict
        self.out[self.radiance_key] = radiance



import astropy.constants as const
import astropy.units as u

import exosim.models.signal as signal
from exosim.tasks.task import Task


class LoadResponsivity(Task):
    """
    Returns the channel responsivity (in counts/Joule) from the quantum efficiency (QE)

    This default implementation loads the quantum efficiency from the input parameters,
    then converts it to responsivity using the relation:

        R(λ) = QE(λ) × λ / (h × c)

    where λ is the wavelength, h is Planck’s constant and c is the speed of light.

    Returns
    --------
    :class:`~exosim.models.signal.Signal`
        channel responsivity

    Raises
    ------
    TypeError:
        if the output is not a :class:`~exosim.models.signal.Signal` class

    UnitConversionError
        if the output has not the right units (counts/Joule)

    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wavelength: :class:`~astropy.units.Quantity`
            wavelength grid.
        time: :class:`~astropy.units.Quantity`
            time grid
        """

        self.add_task_param("parameters", "parameters dict")
        self.add_task_param("wavelength", "wavelength grid")
        self.add_task_param("time", "time grid")

    def execute(self):
        parameters = self.get_task_param("parameters")
        wl = self.get_task_param("wavelength")
        tt = self.get_task_param("time")

        responsivity = self.model(parameters, wl, tt)

        # checking output
        if not isinstance(responsivity, signal.Signal):
            self.error("wrong output format")
            raise TypeError("wrong output format")
        try:
            responsivity.data_units.to(u.m / const.c / const.h * u.count)
        except u.UnitConversionError:
            self.error(
                "{} are not convertible in {}".format(
                    responsivity.data_units,
                    (u.m / const.c / const.h * u.count).unit,
                )
            )
            raise u.UnitConversionError(
                "{} are not convertible in {}".format(
                    responsivity.data_units,
                    (u.m / const.c / const.h * u.count).unit,
                )
            )

        self.debug("responsivity: {}".format(responsivity.data))

        self.set_output(responsivity)

    def model(self, parameters, wavelength, time):
        """

        Parameters
        ----------
        parameters: dict
            dictionary contained the channel parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wavelength: :class:`~astropy.units.Quantity`
            wavelength grid.
        time: :class:`~astropy.units.Quantity`
            time grid.

        Returns
        --------
        :class:`~exosim.models.signal.Signal`
            channel responsivity

        """
        qe_data = parameters["qe"]["data"]
        wl_ = qe_data["Wavelength"]
        qe_ = qe_data[parameters["value"]]
        qe = signal.Dimensionless(data=qe_, spectral=wl_)
        qe.spectral_rebin(wavelength)
        qe.temporal_rebin(time)

        responsivity = signal.Signal(
            spectral=wavelength,
            time=time,
            data=qe.data * wavelength.to(u.m) / const.c / const.h * u.count,
        )
        return responsivity



from copy import deepcopy
from typing import Optional

import astropy.units as u
import numpy as np

from exosim.models.signal import Signal
from exosim.output import Output
from exosim.tasks.task import Task
from exosim.utils import RunConfig
from exosim.utils.checks import check_units
from exosim.utils.iterators import iterate_over_chunks
from exosim.utils.types import ArrayType


class AddGainDrift(Task):
    r"""
    It adds a gain noise map to the array.

    The gain Drift is modeled as a polynomial model for a spectral and time dependent modulation.


    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.
    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        parameters: dict
            channel parameters dictionary
        time: :class:`~astropy.units.Quantity`
            sub-exposures times
        integration_times: :class:`~astropy.units.Quantity`
            sub-exposures integration times
        outputs: :class:`~exosim.output.output.Output` (optional)
            output file
        """

        self.add_task_param("subexposures", "sub-exposures cached signal")
        self.add_task_param("parameters", "channel parameters dictionary")
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("adding gain noise")
        subexposures = self.get_task_param("subexposures")
        parameters = self.get_task_param("parameters")
        output = self.get_task_param("output")

        self.model(subexposures, parameters, output)

    def model(
        self,
        subexposures: Signal,
        parameters: dict,
        output: Optional[Output] = None,
    ) -> None:
        """
        Apply a gain drift model to the subexposures.

        This method models the gain drift as a polynomial
        trend based on time and wavelength.

        Parameters
        ----------
        subexposures : Signal
            Sub-exposures cached signal.
        parameters : dict
            A dictionary containing parameters for the noise model. Expected keys and sub-keys include:
            - 'readout': dict with 'readout_frequency'
            - 'detector': dict with keys 'gain_w0', 'gain_f0', 'gain_coeff_order_t',
                'gain_coeff_t_min', 'gain_coeff_t_max', 'gain_coeff_order_w',
                'gain_coeff_w_min', and 'gain_coeff_w_max'.
        integration_times : np.ndarray
            Sub-exposures integration times.
        output : Optional[Output]
            Output file to write information, defaults to None.

        Returns
        -------
        None
            This method does not return a value but updates the subexposures dataset
            with applied gain noise.

        Notes
        -----
        The method computes Brownian noise based on the frequency parameters in 'parameters' and
        applies a polynomial trend to it. The trend's coefficients are randomly generated within
        specified ranges. The resulting gain noise is then rebinned and applied to the subexposures.
        """

        if output and issubclass(output.__class__, Output):
            out_grp = output.create_group("gain noise")
        else:
            out_grp = None

        time = subexposures.time * u.Unit(subexposures.time_units)
        integration_times = subexposures.metadata["integration_times"]
        if isinstance(integration_times, dict):
            integration_times = integration_times["value"] * u.Unit(
                integration_times["unit"]
            )
            integration_times.to(u.s)

        # Calculating polynomial trend
        order_t = parameters["detector"]["gain_coeff_order_t"]
        coeff_t_min = parameters["detector"]["gain_coeff_t_min"]
        coeff_t_max = parameters["detector"]["gain_coeff_t_max"]
        coeff_t = RunConfig.random_generator.uniform(
            low=coeff_t_min, high=coeff_t_max, size=order_t + 1
        )
        if out_grp is not None:
            out_grp.write_list("gain coeff_t", coeff_t)
        y_t = self._pol_t(time, coeff_t)

        wl_ax = np.arange(0, subexposures.dataset.shape[2], 1)
        order_w = parameters["detector"]["gain_coeff_order_w"]
        coeff_w_min = parameters["detector"]["gain_coeff_w_min"]
        coeff_w_max = parameters["detector"]["gain_coeff_w_max"]
        coeff_w = RunConfig.random_generator.uniform(
            low=coeff_w_min, high=coeff_w_max, size=order_w + 1
        )
        if out_grp is not None:
            out_grp.write_list("gain coeff_w", coeff_w)
        y_w = self._pol_w(wl_ax, coeff_w)

        z = np.zeros([y_t.size, y_w.size])
        for i, y in enumerate(y_t):
            z[i] = y * y_w

        if "gain_drift_amplitude" in parameters["detector"].keys():
            amplitude = parameters["detector"]["gain_drift_amplitude"]
        elif "gain_drift_amplitude_range_min" in parameters["detector"].keys():
            amplitude_min = parameters["detector"][
                "gain_drift_amplitude_range_min"
            ]
            amplitude_max = parameters["detector"][
                "gain_drift_amplitude_range_max"
            ]
            amplitude = RunConfig.random_generator.uniform(
                amplitude_min, amplitude_max
            )
            if out_grp is not None:
                out_grp.write_list(
                    "gain amplitude random_seed", RunConfig.random_seed
                )
        else:
            self.error("missing amplitude definition for gain")

        dinamic_range = z.max() - z.min()
        z *= amplitude / dinamic_range
        z += 1

        if out_grp is not None:
            out_grp.write_array("gain drift", z)
            out_grp.write_array("coeff_t", coeff_t)
            out_grp.write_scalar("amplitude", amplitude)
            out_grp.write_array("coeff_w", coeff_w)

        # Applying the multiplicative noise
        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="applying gain noise"
        ):
            chunk_data = deepcopy(subexposures.dataset[chunk])
            subexposures.dataset[chunk] = (
                chunk_data * z[chunk[0].start : chunk[0].stop, np.newaxis, :]
            )
            subexposures.output.flush()

    @staticmethod
    def _pol_t(tt: np.ndarray, coef: np.ndarray) -> np.ndarray:
        """
        Polynomial trend calculation for time.

        Parameters
        ----------
        tt : np.ndarray
            Time array.
        coef : np.ndarray
            Coefficients for the polynomial trend.

        Returns
        -------
        np.ndarray
            Calculated polynomial trend for time.
        """
        x = ((tt - tt.min()) / (tt.max() - tt.min())).si.value
        y = np.zeros(tt.size)
        for i, c in enumerate(coef):
            y += c * x**i
        return y

    @staticmethod
    def _pol_w(wl: np.ndarray, coef: np.ndarray) -> np.ndarray:
        """
        Polynomial trend calculation for wavelength.

        Parameters
        ----------
        wl : np.ndarray
            spectral axis array [pixel unit].
        coef : np.ndarray
            Coefficients for the polynomial trend.

        Returns
        -------
        np.ndarray
            Calculated polynomial trend for wavelength.
        """
        x = (wl - wl.min()) / (wl.max() - wl.min())
        y = np.zeros(wl.size)
        for i, c in enumerate(coef):
            y += c * x**i
        return y

    @staticmethod
    def _psd(f: np.ndarray, w0: float, f0: float) -> np.ndarray:
        """
        Calculate the power spectral density.

        Parameters
        ----------
        f : np.ndarray
            Frequency array.
        w0 : float
            Parameter for PSD calculation.
        f0 : float
            Parameter for PSD calculation.

        Returns
        -------
        np.ndarray
            Calculated power spectral density.
        """
        psd = np.zeros_like(f)
        psd[1:] = w0 * np.sqrt((f0 / f[1:]) ** 2 + 1)
        return psd

    @staticmethod
    def _noise_generator(
        f: np.ndarray, psd: np.ndarray, out=None
    ) -> np.ndarray:
        """
        Generate noise based on provided frequency and power spectral density.

        Parameters
        ----------
        f : np.ndarray
            Array of frequencies.
        psd : np.ndarray
            Power spectral density corresponding to the frequencies.
        out : object, optional
            Output object to write information, defaults to None.

        Returns
        -------
        np.ndarray
            Generated noise array.
        """
        # Randomize amplitude
        noise = RunConfig.random_generator.normal(
            0, psd * np.sqrt(f[-1]), psd.size
        )
        if out is not None:
            out.write_list("gain amplitude random_seed", RunConfig.random_seed)

        # Randomize phase
        phi = RunConfig.random_generator.normal(-0.5, 1, psd.size) * np.pi
        if out is not None:
            out.write_list("gain phase random_seed", RunConfig.random_seed)

        noisef = noise * np.exp(1j * phi)
        noise = np.fft.irfft(noisef) * np.sqrt(2 * noisef.size - 1)
        return noise



from copy import deepcopy

import numpy as np

from exosim.models.signal import Counts
from exosim.tasks.task import Task
from exosim.utils.checks import check_units
from exosim.utils.iterators import iterate_over_chunks


class ApplySimpleSaturation(Task):
    r"""
    This Task applies a simple model of saturation to pixel counts.
    If the counts in a pixel are higher than the well depth, the counts are set to the well capacity.
    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        parameters: dict
            channel parameters dictionary
        """

        self.add_task_param("subexposures", " ")
        self.add_task_param("parameters", "channel parameters dictionary")

    def execute(self):
        self.info("applying saturation")
        subexposures = self.get_task_param("subexposures")
        parameters = self.get_task_param("parameters")

        self.model(subexposures, parameters)

    def model(self, subexposures: Counts, parameters: dict) -> None:
        """
        Applies the saturation model to the subexposures.

        Parameters
        ----------
        subexposures : Counts
            The subexposures to be saturated.
        parameters : dict
            Dictionary containing saturation parameters like 'detector' and 'well_depth'.

        Returns
        -------
        None
        """

        sat = parameters["detector"]["well_depth"]
        sat = check_units(sat, "ct", force=True).value

        self.info("saturation: {}".format(sat))
        self.info("max counts: {}".format(np.max(subexposures.dataset)))

        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="applying pixel saturation"
        ):
            data = deepcopy(subexposures.dataset[chunk])
            data[data > sat] = sat

            subexposures.dataset[chunk] = data
            subexposures.output.flush()



from abc import ABCMeta, abstractmethod
from typing import Any

import exosim.log as log
from exosim.utils.timed_class import TimedClass


class Task(TimedClass):
    """
    *Abstract class*

    Base class for tasks.
    """

    __metaclass__ = ABCMeta

    _output = None
    _task_input = None
    _task_params = None

    @abstractmethod
    def __init__(self):
        """
        Class initialisation, needed to prepare the task inputs reader
        """
        pass

    @abstractmethod
    def execute(self) -> None:
        """
        Class execution. It runs on call and executes all the task actions returning the outputs.
        It requires the input with correct keywords
        """
        pass

    def __call__(self, **kwargs):
        self.set_log_name()
        self._task_input = kwargs
        self._validate_input_params()
        self._populate_empty_param()
        self.trace("called")
        self.execute()
        self.trace("exited")
        return self.get_output()

    def _validate_input_params(self) -> None:
        for key in self._task_input.keys():
            if key not in self._task_params.keys():
                self.error("Unexpected Task input parameter: {}".format(key))
                raise ValueError

    def _populate_empty_param(self) -> None:
        for key in self._task_params.keys():
            if key not in self._task_input.keys():
                self._task_input[key] = self._task_params[key]["default"]

    def get_output(self) -> Any:
        """Returns the output values."""
        return self._output

    def set_output(self, product: Any) -> None:
        """It sets the values to return."""
        self._output = product

    def get_task_param(self, paramName: str) -> Any:
        """It get the value from the task parameter."""
        return self._task_input[paramName]

    def add_task_param(
        self, param_name: str, param_description: str, default: Any = None
    ) -> None:
        """It adds a parameter for the task."""
        if self._task_params is None:
            self._task_params = {}
        self._task_params[param_name] = {
            "parameters": param_description,
            "default": default,
        }



import numpy as np
from numba import jit

from exosim.output import Output
from exosim.tasks.task import Task
from exosim.utils import RunConfig
from exosim.utils.checks import check_units
from exosim.utils.iterators import iterate_over_chunks


class AddKTC(Task):
    """
    It adds the ktc bias to the sub-exposures.
    This Task produces a new random offset map for each ramp,
    and it adds it to the sub-exposures of the same ramp.
    If an output group is provided, it saves all the random seeds used.

    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        parameters: dict
            channel parameters dictionary
        state_machine: :class:`numpy.ndarray`
            array indicating the exposures number of each sub-exposure.
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        """

        self.add_task_param("subexposures", "sub-exposures cached signal")
        self.add_task_param("parameters", "channel parameters dictionary")
        self.add_task_param("state_machine", "ramp state machine")
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("adding reset bias")
        subexposures = self.get_task_param("subexposures")
        parameters = self.get_task_param("parameters")
        state_machine = self.get_task_param("state_machine")
        output = self.get_task_param("output")

        ktc_std = parameters["detector"]["ktc_sigma"].astype(np.float64)
        ktc_std = check_units(ktc_std, "ct").value

        offset = RunConfig.random_generator.normal(
            0,
            ktc_std,
            (subexposures.dataset.shape[1], subexposures.dataset.shape[2]),
        )

        random_seeds = []

        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="adding ktc noise"
        ):
            if (
                state_machine[chunk[0].start - 1]
                != state_machine[chunk[0].start]
            ):
                offset = RunConfig.random_generator.normal(
                    0,
                    ktc_std,
                    (
                        subexposures.dataset.shape[1],
                        subexposures.dataset.shape[2],
                    ),
                )

            subexposures.dataset[chunk], offset = self.add_offset(
                subexposures.dataset[chunk],
                state_machine[chunk[0]],
                offset,
                ktc_std,
            )
            subexposures.output.flush()
            random_seeds.append(RunConfig.random_seed)

        if output:
            if issubclass(output.__class__, Output):
                out_grp = output.create_group("kTC noise")
                out_grp.write_list("random_seed", random_seeds)
                out_grp.write_array(
                    "chunks_index", np.arange(len(random_seeds))
                )

    @staticmethod
    @jit(nopython=True)
    def add_offset(
        dset: np.ndarray,
        state_machine: np.ndarray,
        offset: np.ndarray,
        bias_std: float,
    ):
        dset = dset.astype(np.float64)
        state_machine = state_machine.astype(np.int64)
        offset = offset.astype(np.float64)

        dset[0] += offset
        for i in range(1, dset.shape[0]):
            if state_machine[i] == state_machine[i - 1]:
                dset[i] += offset
            else:
                offset = np.random.normal(0.0, bias_std, dset[i].shape).astype(
                    np.float64
                )
                dset[i] += offset
        return dset, offset



from copy import deepcopy

import astropy.units as u
import numpy as np

from exosim.models.signal import Signal
from exosim.tasks.task import Task
from exosim.utils.checks import check_units
from exosim.utils.iterators import iterate_over_chunks
from exosim.utils.operations import operate_over_axis
from exosim.utils.types import ArrayType


class AddConstantDarkCurrent(Task):
    """
    It adds constant dark current to all the pixel in the array.
    The dark current is loaded from the parameters

    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.
    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        parameters: dict
            channel parameters dictionary
        integration_times: :class:`~astropy.units.Quantity`
            sub-exposures integration times
        outputs: :class:`~exosim.output.output.Output` (optional)
            output file
        """

        self.add_task_param("subexposures", "sub-exposures cached signal")
        self.add_task_param("parameters", "channel parameters dictionary")
        self.add_task_param(
            "integration_times",
            "subexposures integration times",
        )
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("adding dark current")
        subexposures = self.get_task_param("subexposures")
        parameters = self.get_task_param("parameters")
        integration_times = self.get_task_param("integration_times")
        output = self.get_task_param("output")

        self.model(subexposures, parameters, integration_times, output)

    def model(
        self,
        subexposures: Signal,
        parameters: dict,
        integration_times: ArrayType,
        output=None,
    ) -> None:
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        parameters: dict
            channel parameters dictionary
        integration_times: :class:`~astropy.units.Quantity`
            sub-exposures integration times
        outputs: :class:`~exosim.output.output.Output` (optional)
            output file
        """
        dc = parameters["detector"]["dc_mean"].astype(np.float64)
        dc = check_units(dc, "ct/s")

        self.info("dark current: {}".format(dc))

        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="adding dark current"
        ):
            dc_map = (
                dc.to(u.ct / u.s).value
                * integration_times[chunk[0]].to(u.s).value
            )
            data = deepcopy(subexposures.dataset[chunk])

            subexposures.dataset[chunk] = operate_over_axis(
                data, dc_map, axis=0, operation="+"
            )
            subexposures.output.flush()



import exosim.output as output
from exosim.tasks.task import Task


class LoadPsf(Task):
    """
    It loads the PSFs from files.

    Returns
    -------
    :class:`~numpy.ndarray`
        cube of psfs. axis=0 is time, axis=1 is wavelength, axis=2 is spatial direction, axis=3 is spectral direction.
    :class:`~numpy.ndarray`
        cube normalization factors. It contains the volume of each psf. axis=0 is time, axis=1 is wavelength.

    """

    def __init__(self):
        """

        Parameters
        __________
        parameters: dict
            dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wavelength: :class:`~astropy.units.Quantity`
            wavelength grid.
        time: :class:`~astropy.units.Quantity`
           time grid.
        filename: str
            PSF input file
        output: :class:`~exosim.output.output.Output` (optional)
           output file
        """

        self.add_task_param("wavelength", "wavelength grid")
        self.add_task_param("time", "temporal grid")
        self.add_task_param("parameters", "channel parameters dict")
        self.add_task_param("filename", "PSF input file")
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("loading psf")
        wavelength = self.get_task_param("wavelength")
        time = self.get_task_param("time")
        filename = self.get_task_param("filename")
        parameters = self.get_task_param("parameters")

        psf_cube = self.model(filename, parameters, wavelength, time)
        norms = psf_cube.sum(axis=-1).sum(axis=-1)

        output_file = self.get_task_param("output")
        if output_file and issubclass(output_file.__class__, output.Output):
            output_group = output_file.create_group("psf")
            output_group.write_array(
                "psf_cube",
                psf_cube,
                metadata={
                    "spatial fp_axis": 2,
                    "spectral_fp_axis": 3,
                    "wavelength_axis": 1,
                    "time_axis": 0,
                },
            )
            output_group.write_array("norm", norms)
            output_group.write_quantity("wavelength", wavelength)

        self.set_output([psf_cube, norms])

    def model(self, filename, parameters, wavelength, time):
        """

        Parameters
        ----------
        filename: str
             PSF input file
        parameters: dict
             dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wavelength: :class:`~astropy.units.Quantity`
             wavelength grid.

        Returns
        -------
        :class:`~numpy.ndarray`
            cube of psfs.  axis=0 is time, axis=1 is wavelength, axis=2 is spatial direction, axis=3 is spectral direction.

        """
        raise NotImplementedError



import numpy as np
from astropy import units as u
from astropy.io import ascii

import exosim.models.signal as signal
import exosim.utils.checks as checks
from exosim.tasks.task import Task


class LoadCustom(Task):
    r"""
    Loads a custom SED from a file and scaled it by the solid angle :math:`\pi \left( \frac{R}{D} \right)^2`.

    Returns
    -------
    :class:`~exosim.models.signal.Sed`
        Star Sed
    """

    def __init__(self):
        """
        Parameters
        -----------
        R: :class:`~astropy.units.Quantity` or float
            star radius. If no units are attached is considered as expressed in `m`
        D: :class:`~astropy.units.Quantity` or float
            star distance. If no units are attached is considered as expressed in `m`
        filename: str
            custom sed file path
        """
        self.add_task_param("filename", "custom sed filename")
        self.add_task_param("R", "star radius")
        self.add_task_param("D", "star distannce")

    def execute(self):
        self.info("loading custom sed")
        fname = self.get_task_param("filename")
        R = self.get_task_param("R")
        D = self.get_task_param("D")

        ph = ascii.read(fname, format="ecsv")
        wl_k = checks.find_key(
            ph.keys(), ["Wavelength", "wavelength", "wl"], self
        )
        ph_wl = ph[wl_k].data * ph[wl_k].unit
        sed_k = checks.find_key(ph.keys(), ["Sed", "sed"], self)
        ph_sed = ph[sed_k].data * ph[sed_k].unit

        R = checks.check_units(R, u.m, self)
        D = checks.check_units(D, u.m, self)
        ph_sed *= np.pi * (R / D) ** 2 * u.sr
        print(ph_sed.unit)

        sed = signal.Sed(spectral=ph_wl, data=ph_sed)

        self.debug("custom sed used : {}".format(fname))

        self.set_output(sed)



import warnings

from exosim.tasks.sed import CreatePlanckStar, LoadCustom, LoadPhoenix
from exosim.tasks.task import Task

warnings.filterwarnings("ignore", category=UserWarning, append=True)


class PrepareSed(Task):
    """
    Returns a source SED.
    The SED depends on the source type selected. This Task analyse the inputs and return the desired SED.

    Returns
    -------
    :class:`~exosim.models.signal.Sed`
        Star Sed

    Examples
    ---------
    We first define all the inputs and then we show how to produce the star sed:

    >>> from exosim.tasks.sed import LoadPhoenix
    >>> wavelength = np.linspace(0.5, 7.8, 10000) * u.um
    >>> T = 5778 * u.K
    >>> R = 1 * u.R_sun
    >>> D = 1 * u.au
    >>> M = 1 * u.Msun
    >>> z = 0.0
    >>> g = (cc.G * M.si / R.si ** 2).to(u.cm / u.s ** 2)
    >>> logg = np.log10(g.value)

    To produce a planck star:

    >>> prepareSed = LoadPhoenix()
    >>> sed = PrepareSed(source_type='planck', wavelength=wavelength, T=T, R=R, D=D)

    To load a Phoenix star:
    >>> sed_l = prepareSed(source_type='phoenix', path=phoenix_stellar_model, T=T, R=R, D=D, logg=logg, z=z)

    """

    def __init__(self):
        """
        Parameters
        __________
        source_type: str
            source type, can be `planck`, `phoenix` or `custom`. Default is `planck`
        wavelength: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity` (optional)
            wavelength grid. If no units are attached is considered as expressed in `um`
        R: :class:`~astropy.units.Quantity` or float
            star radius. If no units are attached is considered as expressed in `m`
        D: :class:`~astropy.units.Quantity` or float
            star distance. If no units are attached is considered as expressed in `m`
        filename: str (optional)
            Phoenix file path
        path: str (optional)
            Phoenix directory path
        T: :class:`~astropy.units.Quantity` or float (optional)
            star temperature. If no units are attached is considered as expressed in `K`
        z: float (optional)
            star metallicity.
        logg: float (optional)
            star logG.
        """
        self.add_task_param("source_type", "source type", "planck")
        self.add_task_param("wavelength", "wavelength grid", None)
        self.add_task_param("R", "star radius")
        self.add_task_param("D", "star distance")
        self.add_task_param("T", "star temperature", None)
        self.add_task_param("logg", "star logG", None)
        self.add_task_param("z", "star metallicity", None)
        self.add_task_param("path", "phoenix spectra path", None)
        self.add_task_param("filename", "phoenix file name", None)

    def execute(self):
        self.info("loading source Sed")

        source_type = self.get_task_param("source_type")
        source_type = source_type.lower()
        if source_type == "planck":
            wl = self.get_task_param("wavelength")
            T = self.get_task_param("T")
            R = self.get_task_param("R")
            D = self.get_task_param("D")

            createPlanckStar = CreatePlanckStar()
            sed = createPlanckStar(wavelength=wl, T=T, R=R, D=D)

        elif source_type == "phoenix":
            path = self.get_task_param("path")
            filename = self.get_task_param("filename")
            R = self.get_task_param("R")
            D = self.get_task_param("D")
            T = self.get_task_param("T")
            logg = self.get_task_param("logg")
            z = self.get_task_param("z")

            loadPhoenix = LoadPhoenix()
            sed = loadPhoenix(
                filename=filename, path=path, T=T, R=R, D=D, logg=logg, z=z
            )

        elif source_type == "custom":
            filename = self.get_task_param("filename")
            R = self.get_task_param("R")
            D = self.get_task_param("D")

            loadCustom = LoadCustom()
            sed = loadCustom(filename=filename, R=R, D=D)

        else:
            self.error("not supported source type")
            raise KeyError("not supported source type")

        self.set_output(sed)



from copy import deepcopy

import astropy.units as u
import numpy as np

from exosim.models.signal import Signal
from exosim.tasks.task import Task
from exosim.utils.checks import check_units
from exosim.utils.iterators import iterate_over_chunks
from exosim.utils.operations import operate_over_axis
from exosim.utils.types import ArrayType


class AddDarkCurrentMapNumpy(Task):
    """
    It adds a dark current map to the array.
    The map must be indicated under the `dc_map_filename` keyword.
    The dark current is loaded from a NPY format file (see `numpy documentation <https://numpy.org/devdocs/reference/generated/numpy.lib.format.html>`_).
    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        parameters: dict
            channel parameters dictionary
        integration_times: :class:`~astropy.units.Quantity`
            sub-exposures integration times
        outputs: :class:`~exosim.output.output.Output` (optional)
            output file
        """

        self.add_task_param("subexposures", "sub-exposures cached signal")
        self.add_task_param("parameters", "channel parameters dictionary")
        self.add_task_param(
            "integration_times",
            "subexposures integration times",
        )
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("adding dark current")
        subexposures = self.get_task_param("subexposures")
        parameters = self.get_task_param("parameters")
        integration_times = self.get_task_param("integration_times")
        output = self.get_task_param("output")

        self.model(subexposures, parameters, integration_times, output)

    def model(
        self,
        subexposures: Signal,
        parameters: dict,
        integration_times: ArrayType,
        output=None,
    ) -> None:
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        parameters: dict
            channel parameters dictionary
        integration_times: :class:`~astropy.units.Quantity`
            sub-exposures integration times
        outputs: :class:`~exosim.output.output.Output` (optional)
            output file
        """

        dc_file = parameters["detector"]["dc_map_filename"]
        dc = np.load(dc_file)
        dc = check_units(dc, "ct/s", force=True)

        if dc.shape != subexposures.dataset[0].shape:
            self.error(
                "wrong shape: expected {} but got {}".format(
                    subexposures.dataset[0].shape, dc.shape
                )
            )
            raise OSError()

        self.info("dark current map loaded")

        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="adding dark current"
        ):
            dc_map = (
                dc[np.newaxis, :, :].value
                * integration_times[chunk[0], np.newaxis, np.newaxis]
                .to(u.s)
                .value
            )

            data = deepcopy(subexposures.dataset[chunk])

            subexposures.dataset[chunk] = data + dc_map

            subexposures.output.flush()



from copy import deepcopy

import astropy.units as u
import numpy as np
from numba import jit
from scipy.signal import convolve

import exosim.tasks.instrument as instrument
from exosim.tasks.instrument import CreateIntrapixelResponseFunction
from exosim.tasks.task import Task
from exosim.utils.focal_plane_locations import locate_wavelength_windows
from exosim.utils.iterators import iterate_over_chunks
from exosim.utils.types import ArrayType


class ApplyAstronomicalSignal(Task):
    """
    This task applies the astronomical signal to the sub-exposures.
    To do so, it first convolve the astronomical signal with the instrument line shape (ILS)
    on the focal plane. This is done by populating the focal plane with the ILS and then weighting the contributions to each pixel.
    The resulting model is then convolved with the intrapixel response function (IPRF) and downsampled to the sub-exposure time resolution.
    Then is finally multiplied to the sub-exposure signal.

    Returns
    --------
    :class:`~exosim.models.signal.Counts`
        sub-exposure cached signal class

    """

    def __init__(self) -> None:
        """
        Parameters
        ------------
        source: dict
            source parameters
        model: :class:`np.ndarray`
            model of the astronomical signal in a 2D array (wavelength, time)
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal class
        focal_plane: :class:`~exosim.models.signal.CountsPerSecond`
            channel focal plane
        ils: :class:`np.ndarray`
            instrument line shape. The shape is (time, wavelength, spectral direction)
        timeline : :class:`astropy.units.Quantity`
            timeline to compute the signal
        ch_parameters : dict
            channel parameters
        pointing: (:class:`astropy.units.Quantity`, :class:`astropy.units.Quantity`) (optional)
            telescope pointing direction, expressed ad a tuple of RA and DEC in degrees. Default is ``None``
        source_flux: :class:`~astropy.units.Quantity` (optional)
            source flux on the detector in :math:`ct/s`. Default is ``None``.
        """

        self.add_task_param("source", "source")
        self.add_task_param("model", "astronomical signal model")
        self.add_task_param("subexposures", "subexposures")
        self.add_task_param("focal_plane", "focal_plane")
        self.add_task_param("ils", "instrument line shape")
        self.add_task_param("timeline", "timeline")
        self.add_task_param("ch_parameters", "channel parameters")
        self.add_task_param("pointing", "telescope pointing", None)
        self.add_task_param("source_flux", "source_flux", None)

    def execute(self):
        source = self.get_task_param("source")
        source_flux = self.get_task_param("source_flux")
        model = self.get_task_param("model")
        subexposures = self.get_task_param("subexposures")
        focal_plane = self.get_task_param("focal_plane")
        ils = self.get_task_param("ils")
        timeline = self.get_task_param("timeline")
        ch_parameters = self.get_task_param("ch_parameters")
        pointing = self.get_task_param("pointing")

        integration_times = subexposures.metadata["integration_times"]
        # estimate Subexposure midtimes
        se_time = (
            subexposures.time * subexposures.time_units + integration_times / 2
        )
        self.debug("se mid-times: {}".format(se_time))

        # check if astro timeline is empty
        if timeline.size == 0:
            self.set_output(subexposures)
            self.debug("timeline is empty")
            return
        # extract the timeline starting point from the model
        timeline = timeline.to(se_time.unit)
        start_t = (np.abs(se_time - timeline[0])).argmin()
        end_t = (np.abs(se_time - timeline[-1])).argmin()

        self.debug("model start: {}. model end: {}".format(start_t, end_t))

        # preparing the intrapixel response function
        iprf = self._bild_intrapixel(ch_parameters)
        iprf /= iprf.sum()

        # locate the instrument line shape on the focal plane relative to the source
        _, j0 = locate_wavelength_windows(ils, focal_plane, ch_parameters)
        compute_offset = instrument.ComputeSourcesPointingOffset()
        offset_spectral, _ = compute_offset(
            source=source,
            pointing=pointing,
            parameters=ch_parameters,
        )
        j0 += offset_spectral

        # temporal index of the time dependent ILS
        index = subexposures.metadata["focal_plane_time_indexes"]
        id_start = 0  # index for chunking the model
        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="applying astronomical signal"
        ):
            chunk = list(chunk)  # workaround to replace the tuple first value
            chunk[0] = self.select_chunk_range(chunk[0], start_t, end_t)
            if chunk[0] is not None:
                chunk = tuple(chunk)
                id_stop = id_start + chunk[0].stop - chunk[0].start

                multiplicative_signal, norm = populate(
                    deepcopy(j0),
                    ils,
                    index[chunk[0].start : chunk[0].stop],
                    focal_plane.data.shape[2],
                    model[:, id_start:id_stop],
                    (
                        source_flux / np.sum(source_flux)
                        if source_flux is not None
                        else np.ones(model.shape[0])
                    ),
                )

                # normalise ILS and remove zeroes
                with np.errstate(divide="ignore", invalid="ignore"):
                    multiplicative_signal = np.divide(
                        multiplicative_signal,
                        norm,
                        out=np.ones_like(multiplicative_signal),
                        where=norm != 0,
                    )

                # convolve the model with the intrapixel response function
                for t in range(multiplicative_signal.shape[0]):
                    multiplicative_signal[t] = convolve(
                        multiplicative_signal[t], iprf, mode="same"
                    )
                # downsample the model to the subexposure time resolution
                multiplicative_signal = multiplicative_signal[
                    :, iprf.shape[0] // 2 :: iprf.shape[0]
                ]

                # multiply the model with the subexposure
                subexposures.dataset[chunk] *= multiplicative_signal[
                    :, np.newaxis, :
                ]

                subexposures.output.flush()
                id_start = id_stop
                # asd

        self.set_output(subexposures)

    def select_chunk_range(
        self, chunk: slice, start_t: int, end_t: int
    ) -> slice:
        """
        Selects and adjusts the range of the chunk to be processed based on
        the given start and end times.

        Parameters
        ----------
        chunk : slice
            The original slice object representing the chunk to be processed.
        start_t : int
            The start time to consider for processing.
        end_t : int
            The end time to consider for processing.

        Returns
        -------
        slice
            The adjusted slice object representing the new chunk to be processed.
            Returns None if the chunk is entirely outside the start_t and end_t range.

        Examples
        --------
        >>> select_chunk_range(slice(5, 15, 1), 7, 13)
        slice(7, 13, 1)
        >>> select_chunk_range(slice(5, 15, 1), 16, 20)
        None
        """

        # Check if the chunk intersects with the start time and adjust accordingly
        if start_t > chunk.start and start_t < chunk.stop:
            chunk = slice(start_t, chunk.stop, chunk.step)

        # Check if the chunk intersects with the end time and adjust accordingly
        if end_t < chunk.stop and end_t > chunk.start:
            chunk = slice(chunk.start, end_t, chunk.step)

        # If the chunk is entirely outside the range [start_t, end_t], return None
        if chunk.stop < start_t or chunk.start > end_t:
            return None

        return chunk

    def _bild_intrapixel(self, ch_parameters: dict) -> ArrayType:
        createIntrapixelResponseFunction = CreateIntrapixelResponseFunction()
        iprf, _ = createIntrapixelResponseFunction(parameters=ch_parameters)
        iprf = np.array(iprf)
        # central slice
        # TODO what if the iprf is not odd?
        # TODO what if the wl solution is curved?
        iprf = iprf[iprf.shape[0] // 2, :]
        return iprf


@jit(nopython=True, parallel=True)
def populate(
    j0: np.ndarray,
    psf: np.ndarray,
    psf_index: np.ndarray,
    shape_spectral: int,
    model: np.ndarray,
    weights: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    """Creates the variation of the source signal on the focal plane."""
    n_time = model.shape[1]
    ndrs = len(j0)

    out = np.zeros((n_time, shape_spectral), dtype=np.float64)
    norm = np.zeros((n_time, shape_spectral), dtype=np.float64)

    for k in range(ndrs):
        j0_k = j0[k]
        j1 = j0_k + psf.shape[2]

        j_start = 0
        if j0_k < 0:
            j_start = -j0_k
            j0_k = 0

        j_stop = j_start + psf.shape[2]
        if j1 > shape_spectral:
            j_stop = j_start + shape_spectral - j0_k
            j1 = shape_spectral

        for t in range(n_time):
            psf_index_t = psf_index[t] if psf_index.shape[0] > 1 else 0
            psf_slice = psf[psf_index_t, k, j_start:j_stop]

            contrib = psf_slice * model[k, t] * weights[k]
            out[t, j0_k:j1] += contrib
            norm[t, j0_k:j1] += psf_slice * weights[k]

    return out, norm



from __future__ import annotations

import copy

import astropy.units as u
import numpy as np

import exosim.log as log
import exosim.utils.binning as binning
import exosim.utils.checks as checks
from exosim.models.utils.cachedData import CachedData
from exosim.output.output import Output
from exosim.utils.types import ArrayType, UnitType, ValueType

_invalid_units = "invalid units"


class Signal(log.Logger):
    """
    This class handles data cubes. The cube axes are time (0), spatial (1) and spectral (2) directions.

    Attributes
    -----------
    spectral: :class:`~numpy.ndarray`
        spectral direction grid.
    data: :class:`~numpy.ndarray`
        data table. It has 3 axes: 0. time axis, 1. spatial axis, 2. spectral axis.
    dataset: :class:`h5py.Dataset`
        If cached mode is enabled, it contains the data table. It has 3 axes: 0. time axis, 1. spatial axis, 2. spectral axis.
    time: :class:`~numpy.ndarray`
        time grid.
    spatial: :class:`~numpy.ndarray`
        spatial direction grid.
    spectral_units: str
        define the spectral direction units. Default is `um`
    data_units: str
        define the data units.
    time_units: str
        define the time direction units. Default is `hr`.
    spatial_units: str
        define the spatial direction units. Default is `um`.
    cached: bool
        it tells if cache mode is enabled.
    output: str or  :class:`~exosim.output.hdf5.hdf5.HDF5Output`
        h5py open file used for caching
    dataset_name: :class:`h5py.Dataset`
        h5py dataset used to store the data
    output_path: str
        path where is stored the dataset inside the output file.
    metadata: dict
        signal metadata attached to the class

    Notes
    -----
    To understand caching mode, please look to :class:`~exosim.models.utils.cachedData.CachedData`
    """

    def __init__(
        self,
        spectral: ArrayType = [0] * u.um,
        data: ArrayType = None,
        time: ArrayType = [0] * u.hr,
        data_units: UnitType = None,
        spatial: ArrayType = [0] * u.um,
        shape: tuple[int, int, int] = None,
        cached: bool = False,
        output: str = None,
        output_path: str = None,
        dataset_name: str = None,
        metadata: dict = None,
        dtype: np.dtype = np.float64,
    ) -> None:
        """
        Parameters
        ----------
        spectral: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity` (optional)
            wavelength grid. Must have a single axes. Default is [0] um.
            If the input data is a not :class:`~astropy.units.Quantity`, it's assumed to be expressed in microns,
            otherwhise is converted into microns or pixels.
        data: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity`
            data table. It must have 3 axes: 0. time axis, 1. spatial axis, 2. spectral axis.
            If data is :class:`~astropy.units.Quantity`, data_units are parsed automatically.
        time: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity` (optional)
            time grid. Must have a single axes. Default is [0] hr.
        spatial: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity` (optional)
            spatial direction grid. Must have a single axes.
            If the input data is a not :class:`~astropy.units.Quantity`, it's assumed to be expressed in micron.
            Default is [0] um.
        data_units: str or :class:`~astropy.units.Quantity` (optional)
            define the data units.
            If data is :class:`~astropy.units.Quantity`, data_units are parsed automatically.
        shape: tuple (optional)
            data shape. If data is None, it's used to create a data table with the desired shape.
        cached: bool (optional)
            it enable the cache mode. Default is `False`.
        output: str or :class:`~exosim.output.hdf5.hdf5.HDF5Output` (optional)
            h5 file to use for caching. If `None` a temporary file will be generated. Default is `None`.
        output_path: str (optional)
            path where to store the dataset inside the output file. Deafault is `None`.
        dataset_name: str (optional)
            name to use to store the Dataset into the h5 file named after 'output'. If `None` a random name will be generated. Default is `None`.
        metadata: dict (optional)
            dictionary of metadata.
        dtype: :class:`~numpy.dtype` (optional)
            data type of the data table. Default is `np.float64`.

        """
        self.set_log_name()
        try:
            self.spectral = checks.check_units(
                spectral, "um", self, force=True
            ).value
            self.spectral_units = self._normalize_units(u.um)
        except u.UnitConversionError:
            self.spectral = checks.check_units(spectral, "pix", self).value
            self.spectral_units = self._normalize_units(u.pix)

        self.time = checks.check_units(time, "hr", self, force=True).value
        self.time_units = self._normalize_units(u.hr)

        try:
            self.spatial = checks.check_units(
                spatial, "um", self, force=True
            ).value
            self.spatial_units = self._normalize_units(u.um)
        except u.UnitConversionError:
            self.spatial = checks.check_units(spatial, "pix", self).value
            self.spatial_units = u.pix

        # TODO set up the use of a standard Hdf5Output instead of a common file:
        #  we can use the same file use for output instead of creating a new one
        self.cached = cached
        self.shape = shape
        self.output = output
        self.output_path = output_path
        self.dataset_name = dataset_name

        if data_units is not None:
            self.data_units = self._normalize_units(u.Unit(data_units))
        elif hasattr(data, "unit"):
            self.data_units = self._normalize_units(data.unit)
        else:
            self.data_units = self._normalize_units(u.Unit(""))

        if cached:
            if shape is not None:
                self._data = self._prepare_cached_dataset(
                    fname=output,
                    def_name=dataset_name,
                    data=data,
                    shape=shape,
                    dtype=dtype,
                )
            elif data is not None:
                data = checks.check_units(
                    data, self.data_units, self, force=True
                ).value
                self._data = self._prepare_cached_dataset(
                    fname=output,
                    def_name=dataset_name,
                    data=data,
                    shape=data.shape,
                    dtype=data.dtype,
                )
        else:
            self._data = self._check_data_size(data)
            self._data = checks.check_units(
                self._data, self.data_units, self, force=True
            ).value
        if metadata:
            self.metadata = metadata
        else:
            self.metadata = {}

        self.data = self._data.chunked_dataset[:] if cached else self._data
        self.dataset = self._data.chunked_dataset if cached else None

    def to(self, units: u.Unit) -> None:
        """
        It converts the Signal data into the desired units.

        Parameters
        ----------
        units: :class:`~astropy.units.Unit`
            desired unit

        Examples
        ---------
        >>> import numpy as np
        >>> import astropy.units as u
        >>> from exosim.models.signal import Signal
        >>> wl = np.linspace(0.1, 1, 10) * u.m
        >>> time_grid = np.linspace(1, 5, 10) * u.s
        >>> data = np.random.random_sample((10, 1, 10))*u.m**2
        >>> signal = Signal(spectral=wl, data=data, time=time_grid)
        >>> signal.to(u.cm**2)
        """

        self.data *= self.data_units.to(units)
        self.debug("converted data: {}".format(self.data))
        self.data_units = units

    def _check_data_size(self, data: ArrayType) -> ArrayType:
        while data.ndim < 3:
            data = np.expand_dims(data, axis=0)
        return data

    def _prepare_cached_dataset(
        self,
        fname: str,
        def_name: str,
        data: ArrayType,
        shape: tuple[int, int, int],
        dtype,
    ) -> CachedData:  # type: ignore # noqa: F821
        """
        If cached option is enable, this function initializes a :class:`~exosim.modules.utils.cachedData.CachedData` class.
        """
        cached_data = CachedData(
            axis0=shape[0],
            axis1=shape[1],
            axis2=shape[2],
            output=fname,
            output_path=self.output_path,
            dataset_name=def_name,
            dtype=dtype,
        )
        if data is not None:
            cached_data.chunked_dataset[:] = data

        return cached_data

    @staticmethod
    def _normalize_units(unit: u.Unit) -> u.Unit:
        """Normalize units to a canonical form for consistent comparison."""
        return u.Unit(unit.to_string(format="console"))

    def spectral_rebin(
        self, new_wavelength: ArrayType, fill_value: ValueType = 0.0, **kwargs
    ) -> None:
        """
        It bins the class data over the spectral direction and changes the wavelegth attributes.
        This method is based on :func:`~exosim.utils.binning.rebin`.

        Parameters
        ----------
        new_wavelength: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity`
            new wavelength grid. If no units are attached is considered expressed in 'um'
        fill_value: float or :class:`~astropy.units.Quantity`
            filla value for empry bins. If no units are attached is considered expressed in 'um'.

        Examples
        ----------
        >>> from exosim.models.signal import Signal
        >>> import numpy as np
        >>> import astropy.units as u

        We first define the initial values:

        >>> wavelength = np.linspace(0.1, 1, 10) * u.um
        >>> data = np.ones((10, 1, 10))
        >>> time_grid = np.linspace(1, 5, 10) * u.hr
        >>> signal = Signal(spectral=wavelength, data=data, time=time_grid)
        >>> print(signal.data.shape)
        (10,1,10)

        We now interpolates at a finer wavelength grid:

        >>> new_wl = np.linspace(0.1, 1, 20) * u.um
        >>> signal.spectral_rebin(new_wl)
        >>> print(signal.data.shape)
        (10,1,20)

        We now bin down the to a new wavelength grid:

        >>> signal = Signal(spectral=wavelength, data=data, time=time_grid)
        >>> new_wl = np.linspace(0.1, 1, 5) * u.um
        >>> signal.spectral_rebin(new_wl)
        >>> print(signal.data.shape)
        (10,1,5)

        """
        new_wavelength_ = checks.check_units(
            new_wavelength, "um", self, force=True
        ).value
        self.data = binning.rebin(
            new_wavelength_,
            self.spectral,
            self.data,
            axis=2,
            fill_value=fill_value,
            **kwargs,
        )
        idx = np.where(np.isnan(self.data))
        self.data[idx] = 0.0
        self.spectral = new_wavelength_

    def temporal_rebin(
        self, new_time: ArrayType, fill_value="extrapolate", **kwargs
    ) -> None:
        """
        It bins the class data over the temporal direction and changes the time attributes.
        This method is based on :func:`~exosim.utils.binning.rebin`.


        Parameters
        ----------
        new_time: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity`
            new time grid. If no units are attached is considered expressed in 'hr'

        Examples
        ----------
        >>> from exosim.models.signal import Signal
        >>> import numpy as np
        >>> import astropy.units as u

        We first define the initial values:

        >>> wavelength = np.linspace(0.1, 1, 10) * u.um
        >>> data = np.ones((10, 1, 10))
        >>> time_grid = np.linspace(1, 5, 10) * u.hr
        >>> signal = Signal(spectral=wavelength, data=data, time=time_grid)
        >>> print(signal.data.shape)
        (10,1,10)

        We now interpolates at a finer time grid:

        >>> new_time = np.linspace(1, 5, 20) * u.hr
        >>> signal.temporal_rebin(new_time)
        >>> print(signal.data.shape)
        (20,1,10)

        We now bin down the to a new wavelength grid:

        >>> signal = Signal(spectral=wavelength, data=data, time=time_grid)
        >>> new_time = np.linspace(1, 5, 5) * u.hr
        >>> signal.temporal_rebin(new_time)
        >>> print(signal.data.shape)
        (5,1,10)

        """
        new_time_ = checks.check_units(new_time, "hr", self, force=True).value
        if self.time.size == 1:
            # if a rebin is called on a signal with no time dependence,
            # the data array will be repeated to match the desired size
            a = copy.deepcopy(self.data[0, :, :])
            self.data = np.repeat(a[np.newaxis, :, :], new_time.size, axis=0)

        else:
            # if the rebin is called on a signal with time dependence,
            # the data grid will be rebinned
            self.data = binning.rebin(
                new_time_,
                self.time,
                self.data,
                axis=0,
                fill_value=fill_value,
                **kwargs,
            )

        idx = np.where(np.isnan(self.data))
        self.data[idx] = 0.0
        self.time = new_time_

    # TODO if cached just rename or move. Do not write it again!
    def write(self, output: Output = None, name: str = None) -> None:
        """
        It writes the Signal class into an :class:`~exosim.output.output.Output`.
        The signal class is stored as a dictionary.

        Parameters
        ----------
        output: :class:`~exosim.output.output.Output`
            container to use to write the class

        name: str
            name to use to store

        Examples
        --------
        >>> from exosim.models.signal import Signal
        >>> import numpy as np
        >>> import astropy.units as u

        We first define the initial values:

        >>> wavelength = np.linspace(0.1, 1, 10) * u.um
        >>> data = np.ones((10, 1, 10))
        >>> time_grid = np.linspace(1, 5, 10) * u.hr
        >>> signal = Signal(spectral=wavelength, data=data, time=time_grid)

        Then we store it in a test output HDF5 file

        >>> from exosim.output.hdf5.hdf5 import HDF5Output
        >>> output = os.path.join(test_dir, 'output_test.h5')
        >>> with HDF5Output(output) as o:
        >>>     signal.write(o, 'test_signal')

        Inside the file is now stored a dictionary like the one we can obtain by

        >>> dict(signal)
        """
        if not output:
            output = self.output
        if not name:
            name = self.dataset_name

        if output is None:
            self.warning("No output indicated")
            return

        to_store = dict(self)
        if self.cached:
            to_store.pop("data", None)
        to_store["datatype"] = str(self.__class__)
        output.store_dictionary(to_store, group_name=name)
        self.debug("{} saved".format(name))

    def _find_slice(
        self, start_time: ValueType, end_time: ValueType
    ) -> tuple[int, int]:
        """This function finds a data slice given a time interval."""
        if isinstance(start_time, u.Quantity):
            start_time = start_time.to(u.hr)
        else:
            self.debug("start time assumed to be expressed in hr")
        if isinstance(end_time, u.Quantity):
            end_time = end_time.to(u.hr)
        else:
            self.debug("start time assumed to be expressed in hr")
        start = np.argmin(np.abs(self.time - start_time.value))
        stop = np.argmin(np.abs(self.time - end_time.value))
        return start, stop

    def get_slice(
        self, start_time: ValueType, end_time: ValueType
    ) -> np.ndarray:
        """
        It returnes the data relative to a time slice:

        Parameters
        ----------
        start_time: float or :class:`~astropy.units.Quantity`
            if a float is given, it's assume to be expressed in hours.

        end_time:  float or :class:`~astropy.units.Quantity`
            if a float is given, it's assume to be expressed in hours.

        Returns
        --------
        :class:`~numpy.ndarray`
        """
        start, stop = self._find_slice(start_time, end_time)
        return self.data[start:stop, :, :]

    def set_slice(
        self, start_time: ValueType, end_time: ValueType, data: ArrayType
    ) -> None:
        """
        It replaces the class data values in a specific a time slice:

        Parameters
        ----------
        start_time: float or :class:`~astropy.units.Quantity`
            if a float is given, it's assume to be expressed in hours.

        end_time:  float or :class:`~astropy.units.Quantity`
            if a float is given, it's assume to be expressed in hours.

        data: :class:`~numpy.ndarray`
            data to use as replacement for the existing ones.
        """
        start, end = self._find_slice(start_time, end_time)
        # if self.cached:
        #     self.cached_data.dataset[start:end, :, :] = data
        # else:
        self.data[start:end, :, :] = data

    # Here we define the class iterators. These are usefull to write data

    def __iter__(self):
        for a in ["data", "time", "spectral", "spatial", "metadata", "cached"]:
            yield a, self.__getattribute__(a)
        for a in [
            "data_units",
            "time_units",
            "spectral_units",
            "spatial_units",
        ]:
            yield a, str(self.__getattribute__(a))

    # Here we define the class maths.

    __numpy_ufunc__ = None  # Numpy up to 13.0
    __array_ufunc__ = None  # Numpy 13.0 and above

    def _check_other_val_sum(self, other: ArrayType) -> ArrayType:
        """extracts value and units from the other element and performs the required conversions"""
        if hasattr(other, "unit"):
            val = other.to(self.data_units).value
        elif hasattr(other, "data_units"):
            try:
                val = other.data * other.data_units.to(self.data_units)
            except TypeError:
                val = other._data * other.data_units.to(self.data_units)
        elif hasattr(other, "cached_data"):
            val = other.cached_data
        # elif hasattr(other, 'data'):
        #     val = other.data
        else:
            val = other
        return val

    def _check_other_val_product(self, other: ArrayType) -> ArrayType:
        """extracts value and units from the other element"""
        if hasattr(other, "unit"):
            return other.value, other.unit
        elif hasattr(other, "data_units"):
            return other.data, other.data_units
        else:
            units = u.Unit("")
            val = other
            return val, units

    def _create_new_instance(
        self,
        unit: UnitType = None,
        cached: bool = None,
        shape: tuple[int, int, int] = None,
        output: str | Output = None,
        output_path: str = None,
        metadata: dict = None,
        dataset_name: str = None,
    ) -> Signal:
        # check if overwrite data
        if not cached:
            cached = self.cached
        if not shape:
            shape = self.shape
        if not output:
            output = self.output
        if not output_path:
            output_path = self.output_path
        if not metadata:
            metadata = self.metadata

        # check on cached data
        # if self.cached:
        #     data = self._data.dataset
        # else:
        #
        data = self.data

        # if no unit provided use the first class unit
        if not unit:
            unit = self.data_units

        # given the unit, check the right class to instantiate
        if unit == u.W / u.m**2 / u.um:
            klass = Sed
        elif unit == u.W / u.m**2 / u.um / u.sr:
            klass = Radiance
        elif unit == u.ct / u.s:
            klass = CountsPerSecond
        elif unit == u.ct:
            klass = Counts
        elif unit == u.adu:
            klass = Adu
        elif unit == u.Unit(""):
            klass = Dimensionless
        else:
            klass = Signal
        return klass(
            spectral=self.spectral * self.spectral_units,
            data=data * unit,
            time=self.time * self.time_units,
            spatial=self.spatial * self.spatial_units,
            cached=cached,
            shape=shape,
            dataset_name=dataset_name,
            output=output,
            output_path=output_path,
            metadata=metadata,
        )

    def copy(self, **kwargs) -> Signal:
        """
        It returns a copy of the class.

        Parameters
        ----------
        kwargs: dict
            Dictionary of parameters to overwrite. The paramaters that can be included in the list are `cached`, `metadata`, `dataset_name`, 'output`, `output_path`.

        Returns
        -------
            :class:`~exosim.models.signal.Signal`

        Examples
        ----------

        >>> import numpy as np
        >>> import astropy.units as u
        >>> from exosim.models.signal import Signal
        >>> wl = np.linspace(0.1, 1, 10) * u.um
        >>> data = np.random.random_sample((10, 1, 10))
        >>> time_grid = np.linspace(1, 5, 10) * u.hr
        >>> signal = Signal(spectral=wl, data=data, time=time_grid)
        >>> signal_new = signal.copy()

        """
        return self._create_new_instance(**kwargs)

    def __add__(self, other: ArrayType | Signal | ValueType) -> Signal:
        val = self._check_other_val_sum(other)
        out_class = self._create_new_instance()
        out_class.data = self.data + val
        if isinstance(out_class.data, CachedData):  # type: ignore # noqa: F821
            out_class.cached = True
        return out_class

    def __radd__(self, other: ArrayType | Signal | ValueType) -> Signal:
        return self + other

    def __sub__(self, other):
        val = self._check_other_val_sum(other)
        out_class = self._create_new_instance()
        out_class.data = self.data - val
        if isinstance(out_class.data, CachedData):  # type: ignore # noqa: F821
            out_class.cached = True
        return out_class

    def __rsub__(self, other: ArrayType | Signal | ValueType) -> Signal:
        return (-1) * (self - other)

    def __mul__(self, other: ArrayType | Signal | ValueType) -> Signal:
        val, units = self._check_other_val_product(other)
        out_class = self._create_new_instance(unit=self.data_units * units)
        out_class.data = self.data * val
        if hasattr(out_class.data, "unit"):
            out_class.data = out_class.data.value
        return out_class

    def __rmul__(self, other: ArrayType | Signal | ValueType) -> Signal:
        return self * other

    def __truediv__(self, other: ArrayType | Signal | ValueType) -> Signal:
        val, units = self._check_other_val_product(other)
        out_class = self._create_new_instance(unit=self.data_units / units)
        out_class.data = self.data / val
        if hasattr(out_class.data, "unit"):
            out_class.data = out_class.data.value
        return out_class

    def __rtruediv__(self, other: ArrayType | Signal | ValueType) -> Signal:
        val, units = self._check_other_val_product(other)
        out_class = self._create_new_instance(unit=units / self.data_units)
        out_class.data = val / self.data
        if hasattr(out_class.data, "unit"):
            out_class.data = out_class.data.value
        return out_class

    def __floordiv__(self, other: ArrayType | Signal | ValueType) -> Signal:
        val, units = self._check_other_val_product(other)
        out_class = self._create_new_instance(unit=self.data_units / units)
        out_class.data = self.data // val
        if hasattr(out_class.data, "unit"):
            out_class.data = out_class.data.value
        return out_class

    def __rfloordiv__(self, other: ArrayType | Signal | ValueType) -> Signal:
        val, units = self._check_other_val_product(other)
        out_class = self._create_new_instance(unit=units / self.data_units)
        out_class.data = val // self.data
        if hasattr(out_class.data, "unit"):
            out_class.data = out_class.data.value
        return out_class


class Sed(Signal):
    r"""
    It's a Signal class with data having units of :math:`[W m^{-2} \mu m^{-1}]`
    """

    def __init__(
        self,
        spectral: ArrayType = [0] * u.um,
        data: ArrayType = None,
        time: ArrayType = [0] * u.hr,
        spatial: ArrayType = [0] * u.um,
        *args,
        **kwargs,
    ) -> None:
        self.set_log_name()

        if hasattr(data, "unit") and self._normalize_units(
            data.unit
        ) != self._normalize_units(u.W / u.m**2 / u.um):
            try:
                data = data.to(u.W / u.m**2 / u.um)
            except u.UnitConversionError:
                self.error(_invalid_units)
                raise u.UnitsError(_invalid_units)
        kwargs.pop("data_units", None)

        super().__init__(
            spectral,
            data,
            time,
            data_units=u.W / u.m**2 / u.um,
            spatial=spatial,
            *args,
            **kwargs,
        )


class Radiance(Signal):
    r"""
    It's a Signal class with data having units of :math:`[W m^{-2} \mu m^{-1} sr^{-1}]`
    """

    def __init__(
        self,
        spectral: ArrayType = [0] * u.um,
        data: ArrayType = None,
        time: ArrayType = [0] * u.hr,
        spatial: ArrayType = [0] * u.um,
        *args,
        **kwargs,
    ) -> None:
        self.set_log_name()

        if hasattr(data, "unit") and self._normalize_units(
            data.unit
        ) != self._normalize_units(u.W / u.m**2 / u.um / u.sr):
            try:
                data = data.to(u.W / u.m**2 / u.um / u.sr)
            except u.UnitConversionError:
                self.error(_invalid_units)
                raise u.UnitsError(_invalid_units)
        kwargs.pop("data_units", None)

        super().__init__(
            spectral,
            data,
            time,
            data_units=u.W / u.m**2 / u.um / u.sr,
            spatial=spatial,
            *args,
            **kwargs,
        )


class CountsPerSecond(Signal):
    r"""
    It's a Signal class with data having units of :math:`[ct/s]`
    """

    def __init__(
        self,
        spectral: ArrayType = [0] * u.um,
        data: ArrayType = None,
        time: ArrayType = [0] * u.hr,
        spatial: ArrayType = [0] * u.um,
        *args,
        **kwargs,
    ) -> None:
        self.set_log_name()

        if hasattr(data, "unit") and self._normalize_units(
            data.unit
        ) != self._normalize_units(u.ct / u.s):
            try:
                data = data.to(u.ct / u.s)
            except u.UnitConversionError:
                self.error(_invalid_units)
                raise u.UnitsError(_invalid_units)
        kwargs.pop("data_units", None)

        super().__init__(
            spectral,
            data,
            time,
            data_units=u.ct / u.s,
            spatial=spatial,
            *args,
            **kwargs,
        )


class Counts(Signal):
    r"""
    It's a Signal class with data having units of :math:`[ct]`
    """

    def __init__(
        self,
        spectral: ArrayType = [0] * u.um,
        data: ArrayType = None,
        time: ArrayType = [0] * u.hr,
        spatial: ArrayType = [0] * u.um,
        *args,
        **kwargs,
    ) -> None:
        self.set_log_name()

        if hasattr(data, "unit") and self._normalize_units(
            data.unit
        ) != self._normalize_units(u.ct):
            try:
                data = data.to(u.ct)
            except u.UnitConversionError:
                self.error(_invalid_units)
                raise u.UnitsError(_invalid_units)
        kwargs.pop("data_units", None)

        super().__init__(
            spectral,
            data,
            time,
            data_units=u.ct,
            spatial=spatial,
            *args,
            **kwargs,
        )


class Adu(Signal):
    r"""
    It's a Signal class with data having units of :math:`[adu/s]`
    """

    def __init__(
        self,
        spectral: ArrayType = [0] * u.um,
        data: ArrayType = None,
        time: ArrayType = [0] * u.hr,
        spatial: ArrayType = [0] * u.um,
        *args,
        **kwargs,
    ) -> None:
        self.set_log_name()

        if hasattr(data, "unit") and self._normalize_units(
            data.unit
        ) != self._normalize_units(u.adu):
            try:
                data = data.to(u.adu)
            except u.UnitConversionError:
                self.error(_invalid_units)
                raise u.UnitsError(_invalid_units)
        kwargs.pop("data_units", None)

        super().__init__(
            spectral,
            data,
            time,
            data_units=u.adu,
            spatial=spatial,
            *args,
            **kwargs,
        )


class Dimensionless(Signal):
    """
    It's a Signal class with data having no units
    """

    def __init__(
        self,
        spectral: ArrayType = [0] * u.um,
        data: ArrayType = None,
        time: ArrayType = [0] * u.hr,
        spatial: ArrayType = [0] * u.um,
        *args,
        **kwargs,
    ) -> None:
        self.set_log_name()

        if hasattr(data, "unit") and self._normalize_units(
            data.unit
        ) != self._normalize_units(u.Unit("")):
            data.to(u.Unit(""))
        kwargs.pop("data_units", None)

        super().__init__(
            spectral, data, time, u.Unit(""), spatial=spatial, *args, **kwargs
        )



import numpy as np
from scipy.interpolate import RectBivariateSpline


def fast_convolution(im, delta_im, ker, delta_ker):
    """fast_convolution.
    Convolve an image with a kernel. Image and kernel can be sampled on different
    grids defined.

    Parameters
    __________
    im: :class:`~numpy.ndarray`
        the image to be convolved
    delta_im: float
        image sampling interval
    ker: :class:`~numpy.ndarray`
        the convolution kernel
    delta_ker: float
        Kernel sampling interval
    Returns
    -------
    :class:`~numpy.ndarray`
        the image convolved with the kernel.
    """
    # Fourier transform the kernel
    kerf = np.fft.rfft2(ker)
    ker_k = [
        np.fft.fftfreq(ker.shape[0], d=delta_ker),
        np.fft.rfftfreq(ker.shape[1], d=delta_ker),
    ]
    ker_k[0] = np.fft.fftshift(ker_k[0])
    kerf = np.fft.fftshift(kerf, axes=0)

    # Fourier transform the image
    imf = np.fft.rfft2(im)
    im_k = [
        np.fft.fftfreq(im.shape[0], d=delta_im),
        np.fft.rfftfreq(im.shape[1], d=delta_im),
    ]
    im_k[0] = np.fft.fftshift(im_k[0])
    imf = np.fft.fftshift(imf, axes=0)

    # Interpolate kernel
    kerf_r = RectBivariateSpline(ker_k[0], ker_k[1], kerf.real)
    kerf_i = RectBivariateSpline(ker_k[0], ker_k[1], kerf.imag)
    # Convolve
    imf = imf * (kerf_r(im_k[0], im_k[1]) + 1j * kerf_i(im_k[0], im_k[1]))

    imf = np.fft.ifftshift(imf, axes=0)

    return np.fft.irfft2(imf) * (delta_ker / delta_im) ** 2



import astropy.units as u
import numpy as np
from scipy.interpolate import interp1d

import exosim.models.signal as signal
import exosim.output as output
from exosim.tasks.task import Task
from exosim.utils import RunConfig, check_units


class DarkCurrentMap(Task):
    """
    Produces the channel dark current map

    Returns
    --------
    :class:`~exosim.models.signal.Signal`
        channel dark current map

    Raises
    ------
    TypeError:
        if the output is not a :class:`~exosim.models.signal.Signal` class


    Notes
    -----
    This is a default class with standardized inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        time: :class:`~astropy.units.Quantity`
            time grid
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        """

        self.add_task_param("parameters", "parameters dict")
        self.add_task_param("time", "time grid")
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("creating dark current map")
        parameters = self.get_task_param("parameters")
        tt = self.get_task_param("time")

        dc_map = self.model(parameters, tt)

        # checking output
        if not isinstance(dc_map, signal.Signal):
            self.error("wrong output format")
            raise TypeError("wrong output format")

        self.debug("dark current map: {}".format(dc_map.data))

        output_file = self.get_task_param("output")
        if output_file:
            if issubclass(output_file.__class__, output.Output):
                dc_map.write(output_file, "dc_map")

        self.set_output(dc_map)

    def model(self, parameters, time):
        """

        Parameters
        ----------
        parameters: dict
            dictionary contained the sources parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        time: :class:`~astropy.units.Quantity`
            time grid.

        Returns
        --------
        :class:`~exosim.models.signal.Signal`
            dark current efficiency map

        """

        # variance map
        if "dc_sigma" not in parameters["detector"].keys():
            self.error("dc_sigma not provided in detector configuration")
            raise KeyError("dc_sigma not provided in detector configuration")

        if "dc_mean" in parameters["detector"].keys():
            pass

        elif "dc_median" in parameters["detector"].keys():
            check_units(parameters["detector"]["dc_median"], "ct/s")
            self.compute_dc_mean(parameters["detector"])

        check_units(parameters["detector"]["dc_mean"], "ct/s")
        check_units(parameters["detector"]["dc_sigma"], "ct/s")

        variance_map = RunConfig.random_generator.normal(
            parameters["detector"]["dc_mean"].value,
            parameters["detector"]["dc_sigma"].value,
            (
                1,
                parameters["detector"]["spatial_pix"],
                parameters["detector"]["spectral_pix"],
            ),
        )
        idx = np.where(variance_map <= 0)
        variance_map[idx] = 1e-16
        t = [0] * u.hr
        # temporal variation
        if "dc_aging_factor" in parameters["detector"].keys():
            t0 = parameters["detector"]["dc_aging_time_scale"]
            t0 = check_units(t0, u.hr, force=True)

            # creates the aging effect
            dc_aging = RunConfig.random_generator.normal(
                1,
                parameters["detector"]["dc_aging_factor"],
                (
                    1,
                    parameters["detector"]["spatial_pix"],
                    parameters["detector"]["spectral_pix"],
                ),
            )

            # remove values greater than 1, because QE should become smaller
            dc_aging[dc_aging > 1] = 2 - dc_aging[dc_aging > 1]

            # create the map evolution over time
            variance_map = np.vstack((variance_map, variance_map * dc_aging))
            t = np.hstack((t, t0))
            variance_map = variance_map.reshape(
                (t.size, variance_map.shape[1] * variance_map.shape[2])
            )
            f = interp1d(t, variance_map, fill_value="extrapolate", axis=0)
            variance_map = f(time)
            variance_map = variance_map.reshape(
                (
                    time.size,
                    parameters["detector"]["spatial_pix"],
                    parameters["detector"]["spectral_pix"],
                )
            )
            t = time

        # resized map considering an oversampling factor
        resized_map = variance_map.repeat(
            parameters["detector"]["oversampling"], axis=1
        ).repeat(parameters["detector"]["oversampling"], axis=2)
        dc_map = signal.Dimensionless(
            data=resized_map,
            spectral=np.arange(0, parameters["detector"]["spectral_pix"])
            * u.pix,
            time=t,
        )

        dc_map.temporal_rebin(time)

        return dc_map

    def compute_dc_mean(self, detector):
        """
        Computes the mean of the dark current (dc_mean) from the log-normal distributon.

        Notes
        -----
        The probability density function for the log-normal distributon is:

        .. math::
            pdf(x) = \frac{1}{\\sigma x \\sqrt{2\\pi}}
                      \\exp\\left(-\frac{(\\log(x) - \\mu)^2}{2 \\sigma^2}\right)

        The mean of the pdf can be computed as:

        .. math::
            mean = \\exp(\\mu + \frac{s^2}{2})

        where s is the pdf standard deviation, computed by taking the sqrt of the variance, defined as:

        .. math::

            var = \\left(\\exp(\\sigma^2) - 1\right) \\exp(2 \\mu + \\sigma^2)


        Parameters
        ----------
        detector: dict
            Dictionary for the detector. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`

        Returns
        -------

        None
            Updates the detector dictionary with the value of dc_mean
        """

        mu = np.log(detector["dc_median"].value)

        root = lambda s: detector["dc_sigma"].value ** 2 - (
            np.exp(s**2) - 1
        ) * np.exp(2 * mu + s**2)

        from scipy.optimize import newton

        sigma = newton(func=root, x0=0.5)

        atol = 1e-6
        if not np.isclose(root(sigma), 0.0, atol=atol):
            self.error(
                "dc_sigma root not close to 0 (tol={:.1e})".format(atol)
            )
            raise ValueError(
                "dc_sigma root not close to 0 (tol={:.1e})".format(atol)
            )

        dc_mean = np.exp(mu + sigma**2 / 2)
        detector.update({"dc_mean": dc_mean * u.ct / u.s})
